/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osg/StateSet>
#include <osg/StateAttribute>
#include <osg/Material>
#include <osg/Geometry>
#include <osg/TexMat>
#include <osg/TexEnv>
#include <osg/Texture2D>
#include <osg/LightModel>
#include <osg/Vec3d>
#include <osg/Vec4dH>

#include "RayIntersector.h"
#include "ShadowRay.h"
#include "ReflectedRay.h"
#include "RefractedRay.h"
#include "PhotorealismData.h"

#define M_PI 3.14159265358979323846 // wtf? preco toto nie je definovane, ked je to v math.h ??

// enable certain features
#define PHONG_SHADING
#define TEXTURES
#define REFLECTIONS
#define REFRACTIONS

// enable certain color components
#define AMBIENT
#define DIFFUSE
#define SPECULAR
#define EMISSIVE

// enable light types
#define DIRECTIONAL
#define POINT
#define SPOT

// general macros
#define FACE_FOR_MATERIAL osg::Material::FRONT



using namespace Rayme;


RayIntersector::RayIntersector(const osg::Vec3d& start, const osg::Vec3d& end, double weight, uchar currentDepth, uchar maxDepth)
: osgUtil::LineSegmentIntersector(start, end) {
	init(weight, currentDepth, maxDepth);
}


RayIntersector::RayIntersector(const osg::Vec4d& start, const osg::Vec4d& end, double weight, uchar currentDepth, uchar maxDepth)
: osgUtil::LineSegmentIntersector(start, end) {
	init(weight, currentDepth, maxDepth);
}


RayIntersector::RayIntersector(CoordinateFrame cf, const osg::Vec3d& start, const osg::Vec3d& end, double weight, uchar currentDepth, uchar maxDepth)
: osgUtil::LineSegmentIntersector(cf, start, end) {
	init(weight, currentDepth, maxDepth);
}


RayIntersector::RayIntersector(CoordinateFrame cf, const osg::Vec4d& start, const osg::Vec4d& end, double weight, uchar currentDepth, uchar maxDepth)
: osgUtil::LineSegmentIntersector(cf, start, end) {
	init(weight, currentDepth, maxDepth);
}


RayIntersector::RayIntersector(CoordinateFrame cf, double x, double y, double weight, uchar currentDepth, uchar maxDepth)
: osgUtil::LineSegmentIntersector(cf, x, y) {
	init(weight, currentDepth, maxDepth);
}


void RayIntersector::init(double weight, uchar currentDepth, uchar maxDepth) {
	_weight = weight;
	_currentDepth = currentDepth;
	_maxDepth = maxDepth;
	_color = osg::Vec4d(0, 0, 0, 1);
	_evaluated = false;
	_valid = false;
	_traced = false;
	_reverseDirectionVector = osg::Vec3d(0, 0, 0);
	_debugInfo = false;
	{ // note: the options are managed by the ray manager
		_options.phongShading = true;
		_options.textures = true;
		_options.bilinearTextureFiltering = true;
		_options.recursion = true;
		_options.reflections = true;
		_options.refractions = true;
		_options.reflectionTrimmer = 1.0;
		_options.refractionTrimmer = 1.0;
		_options.selfShadows = true;
	}
}


RayIntersector::~RayIntersector() {
}


RayIntersector* RayIntersector::evaluate() {
	if (_evaluated || !_tracer) {
		return this;
	}
	if (_weight < LIGHT_INTENSITY_CUTOFF) {
		_color.set(0, 0, 0, 1);
		return this;
	}
	if (!_traced) {
		_tracer->traceSingle(this);
	}

	if (_debugInfo) {
		if (_currentDepth == 0) {
			std::cout << std::endl << "------------------------------------------------------------" << std::endl;
		}
		std::cout << "(" << (int) _currentDepth << ") " << std::endl;
		//		std::cout << _start.x() << " " << _start.y() << " " << _start.z() << " -> ";
		//		std::cout << _end.x() << " " << _end.y() << " " << _end.z() << std::endl;
	}

	if (this->containsIntersections() && hasValidIntersections()) {
		/**
		 * The intersection denoting the point hit by the ray.
		 */
		const osgUtil::LineSegmentIntersector::Intersection& intersection = getValidIntersection();
		/**
		 * The Drawable hit by this ray.
		 */
		osg::Drawable *collisionDrawable = intersection.drawable.get();
		/**
		 * StateSet of the Drawable hit by this ray.
		 */
		osg::StateSet *collisionStateSet = collisionDrawable ? collisionDrawable->getStateSet() : NULL;

		// <editor-fold defaultstate="collapsed"  desc="neuspesny pokus o rekurzivne ziskanie StateSet-u">
		/*//TODO neuspesny pokus o ziskanie StateSet-u z predchadzajucich uzlov v NodePath
		if (!collisionStateSet) {
			osg::NodePath::iterator nitr = intersection.nodePath.begin();
			do {
				osg::Geode geode = (*nitr)->asGeode();

				// tu vznika problem, ktory Drawable mam vybrat 
				collisionStateSet = geode ? geode->getDrawable( ? ?)->getStateSet() : (*nitr)->getStateSet();
				nitr++;
			} while (!collisionStateSet && nitr != intersection.nodePath.end());
		}*/
		// </editor-fold>

		// get the base color of the collision point - material and texture

		/**
		 * The material of the object hit by this ray.
		 */
		osg::Material *material = collisionStateSet ? dynamic_cast<osg::Material*> (collisionStateSet->getAttribute(osg::StateAttribute::MATERIAL)) : NULL;
		/**
		 * The material's properties for lighting - ambient, defuse, specular and emissive.
		 */
		osg::Vec4 ka, kd, ks, ke;
		if (material) {
			ka = material->getAmbient(FACE_FOR_MATERIAL);
			kd = material->getDiffuse(FACE_FOR_MATERIAL);
			ks = material->getSpecular(FACE_FOR_MATERIAL);
			ke = material->getEmission(FACE_FOR_MATERIAL);
		} else {
			// the material's properties are unknown - set to OpenGL default
			ka = osg::Vec4(0.2f, 0.2f, 0.2f, 1.0f);
			kd = osg::Vec4(0.8f, 0.8f, 0.8f, 1.0f);
			ks = osg::Vec4(0.0f, 0.0f, 0.0f, 1.0f);
			ke = osg::Vec4(0.0f, 0.0f, 0.0f, 1.0f);
		}
		/**
		 * Color base used to blend with secondary rays.
		 */
		osg::Vec4 diffuseBase = kd;
		/**
		 * The global ambient intensity of the scene.
		 */
		osg::Vec4d globalAmbientIntensity(0, 0, 0, 1);
		{
			osg::LightModel *lightModel = collisionStateSet ? dynamic_cast<osg::LightModel*> (collisionStateSet->getAttribute(osg::StateAttribute::LIGHTMODEL)) : NULL;
			if (lightModel) {
				globalAmbientIntensity = lightModel->getAmbientIntensity();
			} else {
				if (_debugInfo) {
					std::cout << "debug: no global ambient color (light model not set)" << std::endl;
				}
			}
		}
		/**
		 * The Light properties for lighting - ambient, diffuse and specular.
		 */
		osg::Vec4d La, Ld, Ls, current_La, current_Ld, current_Ls;
		/**
		 * Color components for temporary computations
		 */
		double r, g, b, a;
		/**
		 * The position of the currently processed Light.
		 */
		osg::Vec4dH lightPosition;
		/**
		 * The direction of the currently processed Light.
		 */
		osg::Vec3d lightDirection;
		/**
		 * The point of collision.
		 */
		const osg::Vec3d collisionPoint = intersection.getWorldIntersectPoint();
		/**
		 * Normal to the surface at the collision point.
		 * If Phong shading enabled, the value is interpolated later.
		 */
		osg::Vec3d collisionNormal = intersection.getWorldIntersectNormal();
		/**
		 * Shineness of the material hit.
		 */
		const double shininess = material ? material->getShininess(osg::Material::FRONT) : 1;
		/**
		 * Vector from the collision point to the eye. The reverse direction of the ray. Normalized.
		 */
		osg::Vec3d reverseRayDirectionVector = getReverseDirection();
		reverseRayDirectionVector.normalize();


		if (_debugInfo) {
			std::cout << "* " << _start.x() << " " << _start.y() << " " << _start.z();
			std::cout << " T " << collisionPoint.x() << " " << collisionPoint.y() << " " << collisionPoint.z() << std::endl;
			// prints out the actual non-reversed direction
			std::cout << "-> " << -reverseRayDirectionVector.x() << " " << -reverseRayDirectionVector.y() << " " << -reverseRayDirectionVector.z() << std::endl;
		}


		bool textureValid = false;


		if (_debugInfo) {
			std::cout << "depth: " << (int) _currentDepth << std::endl;
			std::cout << "shininess: " << shininess << std::endl;
			std::cout << "ambi " << ka.r() << " " << ka.g() << " " << ka.b() << " " << ka.a() << std::endl;
			std::cout << "diff " << kd.r() << " " << kd.g() << " " << kd.b() << " " << kd.a() << std::endl;
			std::cout << "spec " << ks.r() << " " << ks.g() << " " << ks.b() << " " << ks.a() << std::endl;
		}


		/**
		 * The Geometry hit. Cast from the Drawable.
		 */
		osg::Geometry *collisionGeometry = collisionDrawable ? collisionDrawable->asGeometry() : NULL;
		if (collisionGeometry) {
#ifdef PHONG_SHADING
			if (_options.phongShading) {
				// check the normal binding and if per-vertex normals are used, interpolate the vertex normals (phong interpolation)
				if (collisionGeometry->getNormalBinding() == osg::Geometry::BIND_PER_VERTEX) {
					const osg::Array *a = collisionGeometry->getNormalArray();
					if (a->getType() == osg::Array::Vec3ArrayType) {
						const osg::Vec3Array *drawableNormalArray = dynamic_cast<const osg::Vec3Array*> (a);
						if (drawableNormalArray) {
							osg::Vec3 interpolatedNormal(0, 0, 0);
							osg::Matrix *matrix = intersection.matrix.get();
							osg::Matrix inverse;
							inverse.invert(*matrix);
							for (uint i = 0; i < intersection.indexList.size(); ++i) {
								osg::Vec3d tmp =
									  (*drawableNormalArray)[intersection.indexList[i]] * intersection.ratioList[i];
								tmp = matrix->valid() ? osg::Matrix::transform3x3(inverse, tmp) : tmp;
								interpolatedNormal += tmp;
							}
							collisionNormal = interpolatedNormal;
						}//if drawable normal
					}//if array of Vec3 points
				}//if per vertex normal binding
			}
#endif // phong shading
#ifdef TEXTURES
			if (_options.textures) {
				osg::Vec2d tc(0.f, 0.f); // texture coordinates
				// get texture color at the intersection point
				if (intersection.indexList.size() > 2 && intersection.ratioList.size() == intersection.indexList.size()) {

					osg::Array* texcoords = (collisionGeometry->getNumTexCoordArrays() > 0) ? collisionGeometry->getTexCoordArray(0) : 0;
					osg::Vec2Array* texcoords_Vec2Array = dynamic_cast<osg::Vec2Array*> (texcoords);

					if (texcoords_Vec2Array) {
						// we have tex coord array so now we can compute the final tex coord at the point of intersection.
						for (uint i = 0; i < intersection.indexList.size(); ++i) {
							tc += (*texcoords_Vec2Array)[intersection.indexList[i]] * intersection.ratioList[i];
						}
					}
				}

				osg::TexMat* activeTexMat = 0;
				osg::Texture* activeTexture = 0;
				osg::TexEnv::Mode texEnvMode = osg::TexEnv::MODULATE;
				osg::Vec4 texEnvColor(0, 0, 0, 1);

				if (collisionStateSet) {
					osg::TexMat* texMat = dynamic_cast<osg::TexMat*> (collisionStateSet->getTextureAttribute(0, osg::StateAttribute::TEXMAT));
					if (texMat) {
						activeTexMat = texMat;
					}

					osg::Texture* texture = dynamic_cast<osg::Texture*> (collisionStateSet->getTextureAttribute(0, osg::StateAttribute::TEXTURE));
					if (texture) {
						activeTexture = texture;
					}

					osg::TexEnv * texEnv = dynamic_cast<osg::TexEnv*> (collisionStateSet->getTextureAttribute(0, osg::StateAttribute::TEXENV));
					if (texEnv) {
						texEnvMode = texEnv->getMode();
						texEnvColor = texEnv->getColor();
					}
				}

				if (activeTexMat) {
					osg::Vec4 tc_transformed = osg::Vec4(tc.x(), tc.y(), 0.0f, 0.0f) * activeTexMat->getMatrix();
					tc.x() = tc_transformed.x();
					tc.y() = tc_transformed.y();
				}

				//			std::cout << tc.x() << " " <<tc.y()<<std::endl;

				if (_debugInfo) {
					if (activeTexture) {
						std::cout << "unref textures: " << (activeTexture && activeTexture->getUnRefImageDataAfterApply() ? "yes" : "no") << std::endl;
					} else {
						std::cout << "no texture" << std::endl;
					}
				}

				osg::Image *textureImage = activeTexture ? activeTexture->getImage(osg::Material::FRONT) : NULL;
				if (textureImage) {
					osg::Vec4d textureColor;
					if (_options.bilinearTextureFiltering) {
						textureColor = bilinearTextureFilterColor(textureImage, tc);
					} else {
						textureColor = textureImage->getColor(tc);
					}

					switch (texEnvMode) {
						case osg::TexEnv::MODULATE:
							r = textureColor.r() * kd.r();
							g = textureColor.g() * kd.g();
							b = textureColor.b() * kd.b();
							a = textureColor.a() * kd.a();
							kd.set(CLAMP(r), CLAMP(g), CLAMP(b), CLAMP(a));
							//						std::cout << "modulate" << std::endl;
							break;
						case osg::TexEnv::REPLACE:
							kd.set(textureColor.r(), textureColor.g(), textureColor.b(), textureColor.a());
							//						std::cout << "replace" << std::endl;
							break;
						case osg::TexEnv::DECAL:
							r = (1 - textureColor.r()) * kd.r() + textureColor.r() * texEnvColor.r();
							g = (1 - textureColor.g()) * kd.g() + textureColor.g() * texEnvColor.g();
							b = (1 - textureColor.b()) * kd.b() + textureColor.b() * texEnvColor.b();
							a = textureColor.a() * kd.a();
							kd.set(CLAMP(r), CLAMP(g), CLAMP(b), CLAMP(a));
							//						std::cout << "decal" << std::endl;
							break;
						case osg::TexEnv::BLEND:
							r = (1 - textureColor.a()) * kd.r() + textureColor.a() * textureColor.r();
							g = (1 - textureColor.a()) * kd.g() + textureColor.a() * textureColor.g();
							b = (1 - textureColor.a()) * kd.b() + textureColor.a() * textureColor.b();
							a = textureColor.a() * kd.a();
							kd.set(CLAMP(r), CLAMP(g), CLAMP(b), CLAMP(a));
							//						std::cout << "blend" << std::endl;
							break;
						default:
							break;
					}
					textureValid = true;
				} else {

					//					if (_debugInfo) {
					//						std::cout << "no texture! ";
					//						osg::Image *textureImage = activeTexture ? activeTexture->getImage(osg::Material::BACK) : NULL;
					//						if (textureImage) {
					//							std::cout << "textura ! back";
					//						}
					//
					//						textureImage = activeTexture ? activeTexture->getImage(0) : NULL;
					//						if (textureImage) {
					//							std::cout << "textura 0 ! ";
					//						}
					//
					//						std::cout << std::endl;
					//					}
				}
			}
#endif // textures
		}

		// normalize normal
		collisionNormal.normalize();


		//		La = Ld = Ls = osg::Vec4d(.0f, .0f, .0f, 1.f);
		La = Ld = Ls = osg::Vec4d(.0f, .0f, .0f, .0f);


		// evaluate the light in the collision point using shadow rays iterating through light sources
		for (uint i = 0; i < _tracer->getWorldLightSourceContainer()->sources.size(); i++) {

			lightPosition = osg::Vec4dH(_tracer->getWorldLightSourceContainer()->positions[i]);
			if (!lightPosition.w()) {
				// solves problem with projecting directional lights in infinity
				lightPosition.setWeight(POINT_INFINITY_CORRECTION); //TODO requires a better solution
			}

			/**
			 * Vector running from the light position to the collision point.
			 * Also used for distance measuring. Normalized afterwards.
			 */
			osg::Vec3d collisionToLight = (lightPosition.project() - collisionPoint); // vector from collision point to light

			// other variables for per-light lighting
			double spotCutoff, attenuationFactor, spotAttenuationAngle_cos, spotAttenuationFactor;
			osg::Light *light = _tracer->getWorldLightSourceContainer()->sources[i]->getLight();
			lightDirection = _tracer->getWorldLightSourceContainer()->directions[i]; // should already be normalized
			spotCutoff = light->getSpotCutoff();

			// compute the vector of the reflection
			/**
			 * The vector of the reflection. Used for specular lighting.
			 */
			osg::Vec3d reflectionVector;
			{
				osg::Vec3d lightToCollision = collisionPoint - lightPosition.project();
				lightToCollision.normalize();
				//				reflectionVector = lightToCollision - collisionNormal * (lightToCollision * collisionNormal) * 2;
				//				reflectionVector.normalize();
				reflectionVector = computeReflectionVector(lightToCollision, collisionNormal);
			}

			// NOTE: the intensity factor in each contribution to Ls should be computed as follows:
			// { ((reflectionVector * reverseRayDirectionVector) / (reflectionVector.length() * reverseRayDirectionVector.length()))^shininess },
			// but the vectors are normalized, thus (reflectionVector.length() * reverseRayDirectionVector.length()) == 1
			// and { (reflectionVector * reverseRayDirectionVector)^shininess } is sufficient

			double geometryContribution = 1.0;

			if (!light->getPosition().w()/* == 0.0f*/) {
#ifdef DIRECTIONAL // directional light - sun

				collisionToLight.normalize();
				current_La = light->getAmbient();
				current_Ld = light->getDiffuse() * NONNEGATIVE(collisionNormal * collisionToLight);
				current_Ls = light->getSpecular() * pow(NONNEGATIVE(reflectionVector * reverseRayDirectionVector), shininess);

#endif
			} else if (spotCutoff == 180.0f) {
#ifdef POINT // point light - a bulb

				double lDistance, lDistance_sqr;
				// distance from collision point to light, squared
				lDistance_sqr = collisionToLight.length2();
				// distance from collision point to light
				lDistance = sqrt(lDistance_sqr);
				collisionToLight.normalize();
				// attenuation
				attenuationFactor = (double)1.0 / (double)(light->getConstantAttenuation() + light->getLinearAttenuation() * lDistance + light->getQuadraticAttenuation() * lDistance_sqr);

				if (_debugInfo) {
					std::cout << "att factor: " << attenuationFactor << ", ldistance: " << lDistance << std::endl;
				}

				geometryContribution = NONNEGATIVE(collisionNormal * collisionToLight);
				current_La = (light->getAmbient()) * attenuationFactor;
				current_Ld = light->getDiffuse() * geometryContribution * attenuationFactor; // * lDistance; ??
				if (geometryContribution > 0) {
					current_Ls = light->getSpecular() * pow(NONNEGATIVE(reflectionVector * reverseRayDirectionVector), shininess) * attenuationFactor;
				}

#endif
			} else {
#ifdef SPOT // spot light - a table lamp

				double lDistance, lDistance_sqr;
				// distance from collision point to light, squared
				lDistance_sqr = collisionToLight.length2();
				// distance from collision point to light
				lDistance = sqrt(lDistance_sqr);
				collisionToLight.normalize();

				// distance attenuation factor
				attenuationFactor =  (double)1.0 / (double)(light->getConstantAttenuation() + light->getLinearAttenuation() * lDistance + light->getQuadraticAttenuation() * lDistance_sqr);
				// check wheter the collision point is in the cone of the light
				spotAttenuationAngle_cos = NONNEGATIVE((-collisionToLight) * lightDirection);
//				if (spotAttenuationAngle_cos < cos(spotCutoff * M_PI / 180.0)) {
//					spotAttenuationFactor = 0.f; // no geometryContribution from this light
//				} else {
					spotAttenuationFactor = pow(spotAttenuationAngle_cos, light->getSpotExponent());
//				}
				// combined attenuation
				attenuationFactor *= spotAttenuationFactor;

				if (_debugInfo) {
					std::cout << "att factor: " << attenuationFactor << ", ldistance: " << lDistance << std::endl;
				}

				if (attenuationFactor > 0) {
					geometryContribution = NONNEGATIVE(collisionNormal * collisionToLight);
					current_La = (light->getAmbient()) * attenuationFactor;
					current_Ld = light->getDiffuse() * geometryContribution * attenuationFactor; // * lDistance; ??
					if (geometryContribution > 0) {
						current_Ls = light->getSpecular() * pow(NONNEGATIVE(reflectionVector * reverseRayDirectionVector), shininess) * attenuationFactor;
					}
				}

#endif
			}

			if (geometryContribution >= LIGHT_INTENSITY_CUTOFF) {
				double lightContribution = 1.0;
				{
					// cast a shadow ray to see whether the light is obscured or not
					bool selfShadow = true;
					if (!_options.selfShadows) {
						selfShadow = false;
					}
					ShadowRay shadow(_tracer->getScene(), intersection, lightPosition.project(), selfShadow);
					if (shadow.castsFullShadow()) {
						// light is obscured, this light does not contribute to the color at this point

						//					if ( _debugInfo) {
						//						std::cout << "light: 0%" << std::endl;
						//					}

						continue;
					} else {
						lightContribution = shadow.getLightContribution();
					}
				}

				//			if (_debugInfo) {
				//				std::cout << "light: " << ((lightContribution)*100) << "%" << std::endl;
				//			}


				La += current_La * lightContribution;
				Ld += current_Ld * lightContribution;
				Ls += current_Ls * lightContribution;
			}
		}



		// clamp scene light-source contribution to color components
		La = osg::Vec4d(CLAMP(La.r()), CLAMP(La.g()), CLAMP(La.b()), CLAMP(La.a()));
		Ld = osg::Vec4d(CLAMP(Ld.r()), CLAMP(Ld.g()), CLAMP(Ld.b()), CLAMP(Ld.a()));
		Ls = osg::Vec4d(CLAMP(Ls.r()), CLAMP(Ls.g()), CLAMP(Ls.b()), CLAMP(Ls.a()));

		// calculate the final color of the pixel
		// I = ka * La + kd * Ld + ks * Ls + ke;
		r = 0;
		g = 0;
		b = 0;
		a = diffuseBase.a();
#ifdef AMBIENT
		r += ka.r() * La.r();
		g += ka.g() * La.g();
		b += ka.b() * La.b();
		if (textureValid) {
			// apply to the texture, not just a solid color
			r += kd.r() * globalAmbientIntensity.r();
			g += kd.g() * globalAmbientIntensity.g();
			b += kd.b() * globalAmbientIntensity.b();
		} else {
			r += ka.r() * globalAmbientIntensity.r();
			g += ka.g() * globalAmbientIntensity.g();
			b += ka.b() * globalAmbientIntensity.b();
		}
#endif
#ifdef DIFFUSE
		r += kd.r() * Ld.r();
		g += kd.g() * Ld.g();
		b += kd.b() * Ld.b();
#endif
		r *= a;
		g *= a;
		b *= a;
#ifdef SPECULAR
		r += ks.r() * Ls.r();
		g += ks.g() * Ls.g();
		b += ks.b() * Ls.b();
#endif
#ifdef EMISSIVE
		r += ke.r();
		g += ke.g();
		b += ke.b();
#endif


		// add the light contribution from recursive secondary (reflection and refraction) rays
		if (_options.recursion && _currentDepth < _maxDepth) {
			alterColorBySecondaryRays(r, g, b, diffuseBase, intersection, collisionNormal, shininess);
		}


		// assign the final color, clamp the components to 1.0
		_color = osg::Vec4d(CLAMP(r), CLAMP(g), CLAMP(b), CLAMP(a));


		//		if (_debugInfo) {
		//			std::cout << "color " << _color.r() << " " << _color.g() << " " << _color.b() << std::endl;
		//		}

		//TODO is this validity checking needed still ?
		_valid = true;
	} else {
		if (_debugInfo) {
			std::cout << "missed! ";
			std::cout << "rec.depth: " << (int) _currentDepth << ", s-e: ";
			std::cout << getStart().x() << " " << getStart().y() << " " << getStart().z() << " - " << getEnd().x() << " " << getEnd().y() << " " << getEnd().z() << std::endl;
		}
	}

	_evaluated = true;
	return this;
}


void RayIntersector::alterColorBySecondaryRays(double& r, double& g, double& b,  const osg::Vec4d diffuseBase, 
	  const osgUtil::LineSegmentIntersector::Intersection& intersection, osg::Vec3d collisionNormal, double shininess) {
#if !(defined REFRACTIONS || defined REFLECTIONS)
	return;
#endif
	if (!_options.refractions && !_options.reflections) {
		return;
	}

	// gather photorealism data from the collision node
	PhotorealismData photorealismData;
	for (int i = intersection.nodePath.size() - 1; i >= 0; --i) {
		photorealismData.include(intersection.nodePath[i]);
	}

	osg::Vec3d normalizedRayDirection = getDirection();
	normalizedRayDirection.normalize();


	double reflectance = 0.0;
	double refractance = 0.0;
	double transparency = NONNEGATIVE(CLAMP(1. - diffuseBase.a()));
	double refractiveIndex_n1 = 1.0;
	double refractiveIndex_n2 = 1.0;
	if (transparency >= LIGHT_INTENSITY_CUTOFF) {
		if (photorealismData.contains("Material.refractiveIndex")) {
			// set refractive index manually
			std::stringstream in(photorealismData.get("Material.refractiveIndex"));
			in >> refractiveIndex_n2;
		} else {
			// make up something
			refractiveIndex_n2 = 1.425; // common glass 1.52, water 1.33, averaged
		}

		reflectance = computeReflectanceSchlickApprox(normalizedRayDirection, collisionNormal, refractiveIndex_n1, refractiveIndex_n2);
		refractance = 1 - reflectance;

		if (_debugInfo) {
			std::cout << "n1, n2 " << refractiveIndex_n1 << ", " << refractiveIndex_n2 << std::endl;
			std::cout << "reflectance " << reflectance << std::endl;
			//			std::cout << "normal " << collisionNormal.x() << " " << collisionNormal.y() << " " << collisionNormal.z() << " (" << collisionNormal.length() << ")" << std::endl;
			//			std::cout << "direction " << normalizedRayDirection.x() << " " << normalizedRayDirection.y() << " " << normalizedRayDirection.z() << " (" << normalizedRayDirection.length() << ")" << std::endl;
		}
	}

	double refractionTrimmer = getOptions().refractionTrimmer;
	double reflectionTrimmer = getOptions().reflectionTrimmer;

#ifdef REFRACTIONS
	if (_options.refractions && refractance * _weight >= LIGHT_INTENSITY_CUTOFF) {
		// count the inner refracted-ray vector
		osg::Vec3d refractionVector = computeRefractionVector(normalizedRayDirection, collisionNormal, refractiveIndex_n1, refractiveIndex_n2);

		// cast the secondary ray
		osg::ref_ptr<RefractedRay> secondaryRay = new RefractedRay(refractiveIndex_n2, intersection.getWorldIntersectPoint(), refractionVector, refractance * _weight, _currentDepth + 1, _maxDepth); // note the + 1
		secondaryRay->setTracer(_tracer);
		secondaryRay->setOptions(_options);
		secondaryRay->displayDebugInfo(_debugInfo);
		osg::Vec4d secondaryColor = secondaryRay->getWeightedColor();

		double trans = 1 - diffuseBase.a();

		// shift the secondary color with the material color
		if (trans <= 0.7) {
			r = CLAMP(r + secondaryColor.r() * (1 - diffuseBase.r() * (1 - trans)) * refractionTrimmer);
			g = CLAMP(g + secondaryColor.g() * (1 - diffuseBase.g() * (1 - trans)) * refractionTrimmer);
			b = CLAMP(b + secondaryColor.b() * (1 - diffuseBase.b() * (1 - trans)) * refractionTrimmer);
		} else {
			r = CLAMP(NONNEGATIVE(r + (secondaryColor.r() * (trans) - diffuseBase.g() * (1 - trans) - diffuseBase.b() * (1 - trans)) * refractionTrimmer));
			g = CLAMP(NONNEGATIVE(g + (secondaryColor.g() * (trans) - diffuseBase.r() * (1 - trans) - diffuseBase.b() * (1 - trans)) * refractionTrimmer));
			b = CLAMP(NONNEGATIVE(b + (secondaryColor.b() * (trans) - diffuseBase.r() * (1 - trans) - diffuseBase.g() * (1 - trans)) * refractionTrimmer));
		}

		//		double m;
		//		m = secondaryColor.r();
		//		if (secondaryColor.g() > m) {
		//			m = secondaryColor.g();
		//		}
		//		if (secondaryColor.b() > m) {
		//			m = secondaryColor.b();
		//		}
		//		r = CLAMP(r + m * diffuseBase.r() * trans);
		//		g = CLAMP(g + m * diffuseBase.g() * trans);
		//		b = CLAMP(b + m * diffuseBase.b() * trans);

	}
#endif
#ifdef REFLECTIONS
	if (_options.reflections) {


		osg::Vec4d reflectionRatio(0, 0, 0, 1);
		//		osg::Vec4d diffusionRatio(diffuseBase);

		//TODO reflection intensity
		bool overridden = false;

		if (photorealismData.contains("Material.reflectiveColor")) {
			// set reflection ratio manually
			std::stringstream in(photorealismData.get("Material.reflectiveColor"));
			double r, g, b;
			in >> r >> g >> b;
			reflectionRatio.set(NONNEGATIVE(r), NONNEGATIVE(g), NONNEGATIVE(b), 1);
			overridden = true;

			if (_debugInfo) {
				std::cout << "manual ref.ratio: " << r << " " << g << " " << b << " len:" << reflectionRatio.length() << std::endl;
			}
		} else {
			if (transparency >= LIGHT_INTENSITY_CUTOFF) {
				// reflectance already evaluated
			} else {
				// object not transparent
				if (shininess >= 1.) {
					shininess = shininess / 128.0;
				}
				if (shininess >= 0.5) {
					reflectance = shininess*shininess;
				} else {
					// get from specular
					//TODO toto len v pripade vysokych odrazov
					//					reflectance = specularBase.r() + specularBase.g() + specularBase.b();
					//					reflectance /= 3;
				}


				//				reflectionRatio = specularBase;

			}
			// set reflection ratio from the material characteristics
			//			reflectionRatio = specularBase;
		}

		if ((overridden && reflectionRatio.getVectorPart().length() * _weight > LIGHT_INTENSITY_CUTOFF) || reflectance * _weight > LIGHT_INTENSITY_CUTOFF) {

			// calculate the reflection vector
			osg::Vec3d reflectionVector = computeReflectionVector(normalizedRayDirection, collisionNormal);



			// cast the secondary ray
			osg::ref_ptr<ReflectedRay> secondaryRay = new ReflectedRay(intersection.getWorldIntersectPoint(), reflectionVector, overridden ? _weight : reflectance * _weight, _currentDepth + 1, _maxDepth); // note the + 1
			secondaryRay->setTracer(_tracer);
			secondaryRay->setOptions(_options);
			secondaryRay->displayDebugInfo(_debugInfo);
			osg::Vec4d secondaryColor = secondaryRay->getWeightedColor();

			if (_debugInfo) {
				std::cout << "weight " << _weight << std::endl;
				std::cout << "reflectance " << reflectance << ", overridden: " << (overridden ? "yes" : "no") << std::endl;
				std::cout << "refl.ratio " << reflectionRatio.r() << " " << reflectionRatio.g() << " " << reflectionRatio.b() << std::endl;
				std::cout << "transarency " << transparency << std::endl;
				std::cout << "base      " << r << " " << g << " " << b << std::endl;
				std::cout << "sec.color " << secondaryColor.r() << " " << secondaryColor.g() << " " << secondaryColor.b() << std::endl;
			}

			// reflection secondary light
			if (overridden) {
				r = CLAMP(r + secondaryColor.r() * reflectionRatio.r());
				g = CLAMP(g + secondaryColor.g() * reflectionRatio.g());
				b = CLAMP(b + secondaryColor.b() * reflectionRatio.b());
			} else {
				r = CLAMP(r + secondaryColor.r() * reflectionTrimmer);
				g = CLAMP(g + secondaryColor.g() * reflectionTrimmer);
				b = CLAMP(b + secondaryColor.b() * reflectionTrimmer);
			}

			if (_debugInfo) {
				std::cout << "result    " << r << " " << g << " " << b << std::endl;
			}

			// process the remaining energy of the incomming light - experimental
			//
			//			secondaryColor = secondaryRay->getColor() - secondaryRay->getWeightedColor(); // the remaining incomming light
			//			double attenuation = 1.0 / (1 + 0.8 * secondaryRay->getDistanceToCollision()); //TODO configurable sec.ray attenuation

			//NOTE: remainingLight * M_diffuse * distance_attenuation * ratio_between_specular_and_diffuse
			// diffuse part
			//			r += secondaryColor.r() * texturedSurfaceColor.r() * diffuseBase.a() * attenuation * diffuseBase.r() / (diffuseBase.r() + specularBase.r());
			//			g += secondaryColor.g() * texturedSurfaceColor.g() * diffuseBase.a() * attenuation * diffuseBase.g() / (diffuseBase.g() + specularBase.g());
			//			b += secondaryColor.b() * texturedSurfaceColor.b() * diffuseBase.a() * attenuation * diffuseBase.b() / (diffuseBase.b() + specularBase.b());
			//			// specular part
			//			r += secondaryColor.r() * attenuation * specularBase.r() / (diffuseBase.r() + specularBase.r());
			//			g += secondaryColor.g() * attenuation * specularBase.g() / (diffuseBase.g() + specularBase.g());
			//			b += secondaryColor.b() * attenuation * specularBase.b() / (diffuseBase.b() + specularBase.b());



			//			// compute the diffuse/specular ratio
			//			osg::Vec4d dsRatio(
			//				  cropFactor / (diffusionRatio.r() + reflectionRatio.r()),
			//				  cropFactor / (diffusionRatio.g() + reflectionRatio.g()),
			//				  cropFactor / (diffusionRatio.b() + reflectionRatio.b()),
			//				  1);
			//
			//			// alter the reflection ratio according to reflection angle
			//			// in case the reflection ratio has been manually overriden or the reflection surface is a mirror (1,1,1), do not alter the reflection ratio
			//			if (!overridden && !(reflectionRatio.r() == 1 && reflectionRatio.g() == 1 && reflectionRatio.b() == 1)) {
			//				double cosPhi = reflectionVector * collisionNormal;
			//				reflectionRatio.set(
			//					  reflectionRatio.r() * dsRatio.r(),
			//					  reflectionRatio.g() * dsRatio.g(),
			//					  reflectionRatio.b() * dsRatio.b(),
			//					  1);
			//				reflectionRatio *= (0.3 + 0.7 * (cosPhi < 0 ? 1 + cosPhi : 1 - cosPhi)); // angle coefficient
			//			}
			//			diffusionRatio.set(
			//				  diffusionRatio.r() * dsRatio.r(),
			//				  diffusionRatio.g() * dsRatio.g(),
			//				  diffusionRatio.b() * dsRatio.b(),
			//				  1);
			//			//TODO constant, linear, quadratic
			//			diffusionRatio *= 1.0 / (1 + 0.8 * secondaryRay->getDistanceToCollision()); // distance attenuation
			//
			//
			//			//			if (_debugInfo) {
			//			//				std::cout << "refl.ratio " << reflectionRatio.r() << " " << reflectionRatio.g() << " " << reflectionRatio.b() << std::endl;
			//			//				std::cout << "diff.ratio " << diffusionRatio.r() << " " << diffusionRatio.g() << " " << diffusionRatio.b() << std::endl;
			//			//				std::cout << "base alpha " << baseColor.a() << std::endl;
			//			//			}
			//
			//			// (specular) reflection secondary light
			//			r += reflectionRatio.r() * secondaryColor.r();
			//			g += reflectionRatio.g() * secondaryColor.g();
			//			b += reflectionRatio.b() * secondaryColor.b();
			//
			//#ifdef DIFFUSION_BASED_REFLECTIONS
			//			// diffusion secondary lighting - this should be made blurry
			//			r += diffusionRatio.r() * secondaryColor.r() * baseColor.r();
			//			g += diffusionRatio.g() * secondaryColor.g() * baseColor.g();
			//			b += diffusionRatio.b() * secondaryColor.b() * baseColor.b();
			//#endif
			//
			//			secondaryRay = NULL;
		}
	}

#endif
}


bool RayIntersector::hasValidIntersections() {
	osgUtil::LineSegmentIntersector::Intersections ints = getIntersections();
	osgUtil::LineSegmentIntersector::Intersection intersection;
	osg::Vec3d start = getStart().project();
	bool valid = false;
	for (osgUtil::LineSegmentIntersector::Intersections::iterator it = ints.begin(); it != ints.end(); ++it) {
		intersection = (*it);
		osg::Vec3d ip = intersection.getWorldIntersectPoint();
		if ((start - ip).length2() >= 1e-10) {
			valid = true;
			break; // not equal points, break and store current intersection
		}
	}
	if (valid) {
		_validIntersection = intersection;
	}
	return valid;
}


osgUtil::LineSegmentIntersector::Intersection RayIntersector::getValidIntersection() const {
	return _validIntersection;
}


osg::Vec4d RayIntersector::bilinearTextureFilterColor(const osg::Image *tex, const osg::Vec2d tc) const {
	// cut off the integer part - get the value between 0 and 1
	// then cast the point into pixel units of the texture
	const double u = (tc.x() - (int) tc.x() + (tc.x() < 0 ? 1 : 0)) * tex->s(); // floor
	const double v = (tc.y() - (int) tc.y() + (tc.y() < 0 ? 1 : 0)) * tex->t(); // floor
	// get the bottom left pixel of the quad
	const int x = (int) u; // floor(u)
	const int y = (int) v; // floor(v)

	// cut off the integer part to get the ratios
	const double horizontalRatio = u - x;
	const double verticalRatio = v - y;
	const double horizontalRatio_opposite = 1 - horizontalRatio;
	const double vetricalRatio_opposite = 1 - verticalRatio;

	osg::Vec4d c00, c01, c10, c11;
	c00 = tex->getColor(osg::Vec2d(x, y));
	c01 = tex->getColor(osg::Vec2d(x, (y + 1) % tex->t()));
	c10 = tex->getColor(osg::Vec2d((x + 1) % tex->s(), y));
	c11 = tex->getColor(osg::Vec2d((x + 1) % tex->s(), (y + 1) % tex->t()));

	osg::Vec4d result = (c00 * horizontalRatio_opposite + c10 * horizontalRatio) * vetricalRatio_opposite + (c01 * horizontalRatio_opposite + c11 * horizontalRatio) * verticalRatio;
	return result;
}


osg::Vec3d RayIntersector::getReverseDirection() {
	if (_reverseDirectionVector.x() == 0 && _reverseDirectionVector.y() == 0 && _reverseDirectionVector.z() == 0) {
		_reverseDirectionVector = (osg::Vec4dH(getStart()).project() - osg::Vec4dH(getEnd()).project());
	}
	return _reverseDirectionVector;
}


osg::Vec3d RayIntersector::computeReflectionVector(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal) {
	//NOTE: provided the input vectors are normalized, the output vector will also be
	// ray + 2 * cos_phi * normal = ray + 2 * ( -dot(normal, ray) ) * normal = ...
	return normalizedRayVector - normalizedSurfaceNormal * (normalizedRayVector * normalizedSurfaceNormal) * 2;
}


osg::Vec3d RayIntersector::computeRefractionVector(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal, const double n1, const double n2) {
	//NOTE: provided the input vectors are normalized, the output vector will also be
	const double n = n1 / n2;
	const double cos_phi_i = -(normalizedSurfaceNormal * normalizedRayVector);
	const double sin2_phi_t = n * n * (1. - cos_phi_i * cos_phi_i);
	if (sin2_phi_t > 1.) { // total internal reflection
		return osg::Vec3d(0, 0, 0);
	}
	const double cos_phi_t = sqrt(1. - sin2_phi_t);
	return normalizedRayVector * n + normalizedSurfaceNormal * (n * cos_phi_i - cos_phi_t);
}


double RayIntersector::computeReflectanceSchlickApprox(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal, const double n1, const double n2) {
	double r0 = (n1 - n2) / (n1 + n2);
	r0 *= r0; // r0^2
	double cos_phi = -(normalizedSurfaceNormal * normalizedRayVector);
	if (n1 > n2) {
		const double n = n1 / n2;
		const double sin2_phi = n * n * (1. - cos_phi * cos_phi);
		if (sin2_phi > 1.) { // total internal reflection
			return 0;
		}
		cos_phi = sqrt(1. - sin2_phi);
	} else {
		if (cos_phi < 0) {
			cos_phi = -cos_phi;
		}
	}
	const double x = 1. - cos_phi;
	return r0 + (1. - r0) * x * x * x * x * x; // (1.0 - cos_phi)^5
}


osg::Vec3d RayIntersector::getDirection() {
	return -getReverseDirection();
}


RayIntersector * RayIntersector::setTraced(bool traced) {
	_traced = traced;
	return this;
}


void RayIntersector::displayDebugInfo(bool display) {
	_debugInfo = display;
}


double RayIntersector::getDistanceToCollision() const {
	osg::Vec3d collisionPoint = getValidIntersection().getWorldIntersectPoint();
	osg::Vec3d distv(collisionPoint - getStart().getVectorPart());
	return distv.length();
}


double RayIntersector::getDistanceToCollision2() const {
	osg::Vec3d collisionPoint = getValidIntersection().getWorldIntersectPoint();
	osg::Vec3d distv(collisionPoint - getStart().getVectorPart());
	return distv.length2();
}


void RayIntersector::setOptions(const Options o) {
	_options = o;
}


const RayIntersector::Options& RayIntersector::getOptions() const {
	return _options;
}

