/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#ifndef RAYME_RAYINTERSECTOR_H
#define	RAYME_RAYINTERSECTOR_H


#include <math.h>
#include <osgUtil/LineSegmentIntersector>
#include <osg/Vec4dH>
#include <osg/Vec3d>

#include "Tracer.h"
#include "common.h"


#define POINT_PRECISION 1e-10 // offset of the starting point of a secondary ray or the threshold for two points being equal
#define POINT_INFINITY 1e+10 // offset of the ending point of a secondary ray
#define POINT_INFINITY_CORRECTION 1e-100 // if the light's position weight is set to 0 (directional light)
#define LIGHT_INTENSITY_CUTOFF 1e-2 // if a light's contribution is lower then this, the light is ignored


namespace Rayme {
	class Tracer;
	//
	/**
	 * 
	 */
	class RayIntersector : public osgUtil::LineSegmentIntersector {
	public: // constructors & destructors
		RayIntersector(const osg::Vec3d& start, const osg::Vec3d& end, double weight, uchar currentDepth, uchar maxDepth);
		RayIntersector(const osg::Vec4d& start, const osg::Vec4d& end, double weight, uchar currentDepth, uchar maxDepth);
		RayIntersector(CoordinateFrame cf, const osg::Vec3d& start, const osg::Vec3d& end, double weight, uchar currentDepth, uchar maxDepth);
		RayIntersector(CoordinateFrame cf, const osg::Vec4d& start, const osg::Vec4d& end, double weight, uchar currentDepth, uchar maxDepth);
		RayIntersector(CoordinateFrame cf, double x, double y, double weight, uchar currentDepth, uchar maxDepth);
		virtual ~RayIntersector();
		struct Options {
			bool phongShading;
			bool textures;
			bool bilinearTextureFiltering;
			bool recursion;
			bool reflections;
			bool refractions;
			bool selfShadows;
			double reflectionTrimmer;
			double refractionTrimmer;
		};

	public: // methods
		/**
		 * This method evaluates ray intersections gathered in the scene.
		 * If the ray has not been previously traced in the scene, this method invokes the tracing.
		 * This methd needs not to be called directly, as it is called upon calling getColor().
		 * @see getColor()
		 */
		virtual RayIntersector* evaluate();
		/**
		 * This method was originally provided to unleash the power of double dispatch. No longer needed.
		 * @todo remake to single dispatch.
		 */
		virtual void selfTrace(Tracer const& tracer) = 0;
		/**
		 * Returns reversed ray direction.
		 * Reversed direction as compared to the ray-tracing ray terminology.
		 * The vector pointing into the eye/camera, the direction of natural photon trajectory (for primary rays).
		 */
		virtual osg::Vec3d getReverseDirection();
		/**
		 * Returns ray direction in the ray-tracing terminology.
		 * The diretion pointing from the eye/camera into the scene (for primary rays).
		 * Reversed, compared to the natural photon trajectory direction.
		 */
		virtual osg::Vec3d getDirection();
		/**
		 * Indicate, that the ray has been traced in the scene.
		 */
		RayIntersector* setTraced(bool traced = true);
		/**
		 * Get the distance from start to collision. Squared.
		 * NOTE: precede with isValid() method
		 */
		double getDistanceToCollision2() const;
		/**
		 * Get the distance from start to collision.
		 * NOTE: precede with isValid() method
		 */
		double getDistanceToCollision() const;
		void setOptions(const Options o);
		const Options& getOptions() const;
		/**
		 * Computes the vector of reflection.
		 * @param normalizedRayVector 
		 */
		static osg::Vec3d computeReflectionVector(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal);
		static osg::Vec3d computeRefractionVector(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal, const double n1, const double n2);
		static double computeReflectanceSchlickApprox(osg::Vec3d const& normalizedRayVector, osg::Vec3d const& normalizedSurfaceNormal, const double n1, const double n2);


	protected: // methods
		osg::Vec4d bilinearTextureFilterColor(const osg::Image *tex, const osg::Vec2d tc) const;
		virtual void alterColorBySecondaryRays(double& r, double& g, double& b, const osg::Vec4d diffuseBase,
											const osgUtil::LineSegmentIntersector::Intersection& intersection, osg::Vec3d collisionNormal, double shininess);
		virtual bool hasValidIntersections();
		osgUtil::LineSegmentIntersector::Intersection getValidIntersection() const;

	private: // methods
		void init(double weight, uchar currentDepth, uchar maxDepth);

	protected: // members
		uchar _currentDepth;
		uchar _maxDepth;
		double _weight;
		osg::Vec4d _color;
		Tracer* _tracer;
		bool _traced;
		bool _evaluated;
		bool _valid;
		osg::Vec3d _reverseDirectionVector;
		osgUtil::LineSegmentIntersector::Intersection _validIntersection;
		Options _options;

	public: // inline methods
		/**
		 * Method retrieves the computed color intensity of the ray.
		 * Calls evaluate() if this has not been previously called.
		 * @see evaluate()
		 * NOTE: precede with evaluate() and isValid() method combo to avoid confusion
		 */
		virtual inline osg::Vec4d getColor() {
			if (!_evaluated) {
				evaluate();
			}
			return _color;
		}
		virtual inline osg::Vec4d getWeightedColor() {
			return getColor() * _weight;
		};
		inline void setWeight(double weight) {
			_weight = weight;
		};
		inline double getWeight() const {
			return _weight;
		};
		/**
		 * Set the maximum recursion depth. 
		 * The depth can also be set using a constructor.
		 */
		inline RayIntersector* setMaxRecursionDepth(uchar depth) {
			_maxDepth = depth;
			return this;
		}
		/**
		 * Set current recursion depth. 
		 * The depth can also be set using a constructor.
		 */
		inline RayIntersector* setCurrentRecursionDepth(uchar depth) {
			_currentDepth = depth;
			return this;
		}
		/**
		 * Disable recursion. No consequent rays will be created.
		 * Can also be used to enable recursion again.
		 */
		inline RayIntersector * disableRecursion(bool disable = true) {
			_options.recursion = !disable;
			return this;
		}
		/**
		 * Set the tracer object.
		 */
		inline void setTracer(Tracer* tracer) {
			_tracer = tracer;
		}
		/**
		 * Checks whether the ray is valid.
		 * A ray is valid, if it has been evaluated and contains a valid intersection.
		 */
		inline bool isValid() const {
			return _evaluated && _valid;
		}
		/**
		 * Checks whether the ray has been evaluated.
		 */
		inline bool isEvaluated() const {
			return _evaluated;
		}

	public: // debugging
		bool _debugInfo;
		/**
		 * Method for debugging purposes, mainly for point-tracing information print-out.
		 */
		void displayDebugInfo(bool display = true);

	};

}

#endif	/* RAYME_RAYINTERSECTOR_H */

