/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osg/Node>
#include <osgUtil/IntersectionVisitor>
#include <osgUtil/LineSegmentIntersector>
#include <osg/StateSet>
#include <osg/StateAttribute>
#include <osg/Material>
#include <osgDB/WriteFile>
#include <osg/MatrixTransform>

#include "RayManager.h"
#include "LightSourceVisitor.h"
#include "PrimaryRay.h"
#include "TracerThread.h"

// setup macros
#define THREADS_PER_PROCESSOR 2
#define DEFAULT_MAX_RECUSION_DEPTH 3
#define DEFAULT_SAMPLES_PAR_PIXEL 1
#define DEFAULT_CHUNK_SIZE 100

// behaviour macros
//#define PRINT_LIGHT_SOURCES // print out info about light sources in the scene
//#define PRINT_THREAD_INFO // thread will print info about its current state
#define NOSCENECOPY // the base render image is blank, otherwise the scene window image is used
#define DISPLAY_PROGRESSIVE // displays the intermediate results of the computation
//#define RANDOM_DYNAMIC_RAY_DISTRIBUTION // when dynamically distributing rays, this will force random ray order

#ifdef RANDOM_DYNAMIC_RAY_DISTRIBUTION
#include <algorithm>
#include <functional>
#include <ctime>
#endif


using namespace Rayme;


RayManager::RayManager(osg::Camera* camera, osg::Node* scene, osg::Image* outputBuffer, osg::Image* sceneSnapshot, osgViewer::View* renderView) :
_active(false),
_scene(scene), _masterCamera(camera), _renderMatrixInverse(new osg::RefMatrix), _renderView(renderView), _outputBuffer(outputBuffer),
_sceneSnapshot(sceneSnapshot), _distribution(RayManager::DYNAMIC), _samplesPerPixel(DEFAULT_SAMPLES_PAR_PIXEL), _recursionDepth(DEFAULT_MAX_RECUSION_DEPTH),
_threadCount(0), _threadsPerProcessor(THREADS_PER_PROCESSOR), _chunkSize(DEFAULT_CHUNK_SIZE),
_pointTracing(false), _pointTracer(NULL) {
	_processorCount = OpenThreads::GetNumberOfProcessors();
	_referenceOptions.phongShading = true;
	_referenceOptions.textures = true;
	_referenceOptions.bilinearTextureFiltering = true;
	_referenceOptions.recursion = true;
	_referenceOptions.reflections = true;
	_referenceOptions.refractions = true;
	_referenceOptions.selfShadows = true;
	_referenceOptions.reflectionTrimmer = 1.0;
	_referenceOptions.refractionTrimmer = 1.0;
	_printProgressToConsoleThreshold = -1;
}


RayManager::~RayManager() {
	cancel();
}


void RayManager::updateRenderMatrixInverse() {
	if (_masterCamera.get()) {
		osg::Matrixd matrix;
		matrix.preMult(_masterCamera->getViewport()->computeWindowMatrix());
		matrix.preMult(_masterCamera->getProjectionMatrix());
		matrix.preMult(_masterCamera->getViewMatrix());
		_renderMatrixInverse->invert(matrix);
	}
}


void RayManager::setRenderMatrixInverse(osg::Matrixd& matrix) {
	_renderMatrixInverse->invert(matrix);
}


void RayManager::requestNewFrame() {
	stopPointTracing();
	if (!hasDedicatedThreads()) {
		this->run();
	} else {
		this->startThread();
	}
}


void RayManager::run() {
	if (!isActive()) {
		activate();
		setProgress(0);

#ifdef PRINT_THREAD_INFO
		std::cout << "ray manager thread active" << std::endl;
#endif
		YieldCurrentThread();

		// measure overall duration of frame generation
		osg::Timer_t frameStartTick = osg::Timer::instance()->tick();

		std::cout << "Rendering at " << _outputBuffer->s() << "x" << _outputBuffer->t() << std::endl;

		// run the ray tracing computation
		if (hasDedicatedThreads()) {
			startSceneRayTracing(_distribution, _threadCount ? _threadCount : _threadsPerProcessor * _processorCount);
		} else {
			startSceneRayTracing(RayManager::STATIC, 0);
		}

		//		startSceneRayTracing((RayDistributionModel) 12, 0);

		// force the GPU to reload the updated texture
		_outputBuffer->dirty();
		// request the ray display to redraw
		if (_renderView.get()) {
			_renderView->requestRedraw();
		}

#ifdef PRINT_THREAD_INFO
		std::cout << "ray manager thread finished";
#endif

		osg::Timer_t frameEndTick = osg::Timer::instance()->tick();
		osg::Timer_t delta = osg::Timer::instance()->delta_s(frameStartTick, frameEndTick);

#ifdef PRINT_THREAD_INFO
		std::cout << " in " << delta << " seconds." << std::endl;
#else
		std::cout << "Frame in " << delta << " seconds." << std::endl;
#endif
		_lastRuntime = delta;

		std::cout << std::endl;
		deactivate();
	} else {
#ifdef PRINT_THREAD_INFO


		std::cout << "ray manager thread being cancelled ... " << std::endl;
#endif
		cancel();
#ifdef PRINT_THREAD_INFO
		std::cout << "ray manager thread cancelled" << std::endl;
#endif
		deactivate();
	}
}


void RayManager::initThreads(const uint threadCount) {
	// collect light sources in the scene
	WorldLightSourceContainer * lc = new WorldLightSourceContainer();
	collectLightSources(*lc);


	// create ray tracing objects (threads)
	_tracerThreads.clear();
	for (uint i = 0; i < threadCount; i += 1) {
		_tracerThreads.push_back(new TracerThread(i, _scene.get(), lc));
	}
}


void RayManager::startSceneRayTracing(RayManager::RayDistributionModel distribution, const uint threadCount) {
	updateRenderMatrixInverse();
	initThreads(threadCount ? threadCount : 1);
	prepareOutputImage();

	const ushort width = _outputBuffer->s();
	const ushort height = _outputBuffer->t();

	switch (distribution) {
		case RayManager::STATIC:
			// create primary rays, assign them to certain tracing objects
			for (ushort x = 0; x < width; x += 1) {
				for (ushort y = 0; y < height; y += 1) {
					createAndAssignBeam(_tracerThreads[(width * y + x) % threadCount], x, y);
				}
			}

#ifdef PRINT_THREAD_INFO
			for (uint i = 0; i < threadCount; i += 1) {
				std::cout << i << " / " << _tracerThreads[i]->getRays()->getIntersectors().size() << std::endl;
			}
#endif

			// start the computation ray-tracing threads
			if (threadCount) {
				// issue parallel computation
				uint totalRayCount = 0;
				for (uint i = 0; i < _tracerThreads.size(); i += 1) {
					_tracerThreads[i]->startThread();
					totalRayCount += _tracerThreads[i]->getRayCount();
				}

				// wait for the rt threads to finish, meanwhile evaluate overal progress
#ifdef PRINT_THREAD_INFO
				std::cout << "ray manager thread waiting for tracer threads to finish" << std::endl;
#endif
				uint countFinishedThreads;
				uint totalFinishedRays = 0;
				do {
					totalFinishedRays = countFinishedThreads = 0;
					OpenThreads::Thread::microSleep(200000); // 5 times a second is enough
					for (uint i = 0; i < _tracerThreads.size(); i += 1) {
						countFinishedThreads += _tracerThreads[i]->isRunning() ? 0 : 1;
						totalFinishedRays += _tracerThreads[i]->getFinishedCount();
					}
					setProgress(totalFinishedRays * 100 / totalRayCount);
				} while (countFinishedThreads < _tracerThreads.size() && isActive());
			} else {
				// issue sequential computation
				for (uint i = 0; i < _tracerThreads.size(); i += 1) {
					_tracerThreads[i]->trace();
				}
			}

			// write the results
			for (uint i = 0; i < _tracerThreads.size(); i += 1) {
				_tracerThreads[i]->writeImage(_outputBuffer);
			}
			break;
		case RayManager::DYNAMIC:
			// NOTE: parallel computation only
		{
			uint totalRayCount = width * height * _samplesPerPixel, totalFinishedRays = 0, currentRayIndex = 0;
			ushort x, y;

#ifdef DISPLAY_PROGRESSIVE
			uchar lastProgress = getProgress();
			osg::Timer_t tick = osg::Timer::instance()->tick();
#endif
#ifdef RANDOM_DYNAMIC_RAY_DISTRIBUTION
			srand(unsigned ( time(NULL)));
			std::vector<uint> indexVector;
			for (uint i = 1; i < totalRayCount; ++i) {
				indexVector.push_back(i);
			}
			for (uint i = 0; i < 4; ++i) {
				std::random_shuffle(indexVector.begin(), indexVector.end());
			}
#endif
			uchar lastPrintout = 0;
			do {
				for (uint i = 0; i < _tracerThreads.size(); i += 1) {
					// check whether the thread is currently running
					if (!_tracerThreads[i]->isRunning()) {
						// write the computed data and reset the tracer, if any ray assigned
						if (_tracerThreads[i]->getRayCount()) {
							_tracerThreads[i]->writeImage(_outputBuffer);
							totalFinishedRays += _tracerThreads[i]->getRayCount();
							_tracerThreads[i]->reset();
						}

#ifdef PRINT_THREAD_INFO
						std::cout << "rays " << currentRayIndex << " to " << (currentRayIndex + _chunkSize <= totalRayCount ? currentRayIndex + _chunkSize : totalRayCount) << " assigned to thread #" << i << std::endl;
#endif

						for (uint limit = currentRayIndex + _chunkSize <= totalRayCount ? currentRayIndex + _chunkSize : totalRayCount;
							  currentRayIndex < limit; currentRayIndex += 1) {
#ifdef RANDOM_DYNAMIC_RAY_DISTRIBUTION
							x = indexVector[currentRayIndex] % width;
							y = indexVector[currentRayIndex] / width;
#else
							x = currentRayIndex % width;
							y = currentRayIndex / width;
#endif
							createAndAssignBeam(_tracerThreads[i], x, y);
						}
						if (isActive()) {
							_tracerThreads[i]->startThread();
						}
					}
				}
				uchar tmpProgress = totalFinishedRays * 100 / totalRayCount;
				setProgress(tmpProgress);
				if (_printProgressToConsoleThreshold >= 0) {
					if ((tmpProgress >= lastPrintout + _printProgressToConsoleThreshold)) {
						std::cout << "Progress: " << (int) tmpProgress << "%" << std::endl;
						lastPrintout = tmpProgress;
					}
				}
#ifdef DISPLAY_PROGRESSIVE
				if (tmpProgress > lastProgress && osg::Timer::instance()->delta_s(tick, osg::Timer::instance()->tick()) >= 1) {
					_outputBuffer->dirty();
					if (_renderView.get()) {
						_renderView->requestRedraw();
					}
					lastProgress = getProgress();
					tick = osg::Timer::instance()->tick();
				}
#endif

				OpenThreads::Thread::microSleep(20000); // 50 times a second is enough
			} while (totalFinishedRays < totalRayCount && isActive());
		}
			break;
		default:
			std::cout << "WARNING! single thread being used."; //TODO for profiling only

			uint totalRayCount = width * height * _samplesPerPixel, totalFinishedRays = 0, currentRayIndex = 0;
			ushort x, y;
			do {
				for (uint limit = currentRayIndex + _chunkSize <= totalRayCount ? currentRayIndex + _chunkSize : totalRayCount;
					  currentRayIndex < limit; currentRayIndex += 1) {
					x = currentRayIndex % width;
					y = currentRayIndex / width;
					createAndAssignBeam(_tracerThreads[0], x, y);
				}
				_tracerThreads[0]->run();
				if (_tracerThreads[0]->getRayCount()) {
					_tracerThreads[0]->writeImage(_outputBuffer);
					totalFinishedRays += _tracerThreads[0]->getRayCount();
					_tracerThreads[0]->reset();
				}
				//				totalFinishedRays += _tracerThreads[0]->getRayCount();
			} while (totalFinishedRays < totalRayCount);

			break;
	}


}


void RayManager::createAndAssignBeam(Tracer* tracer, const ushort x, const ushort y, bool debugInfo) {
	osg::Vec4dH start, end;
	PrimaryRay *ray;
	double OpenGL_offset_x, OpenGL_offset_y, offset_x, offset_y;
	OpenGL_offset_x = OpenGL_offset_y = 0.5;
	ushort maxRecursionDepth = _recursionDepth;
	uchar samplesPerPixel = _samplesPerPixel;

	switch (samplesPerPixel) {
		case 4: // supersampling AA (multisampling) 4x (rotated grid)
			offset_x = 0.1;
			offset_y = 0.4;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = -0.4;
			offset_y = 0.1;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = -0.1;
			offset_y = -0.4;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = 0.4;
			offset_y = -0.1;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);
			break;
		case 8: // supersampling AA (multisampling) 4x (rotated grid)
			offset_x = -.45;
			offset_y = .45;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = .1125;
			offset_y = .3375;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = .45;
			offset_y = .225;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = .225;
			offset_y = .1125;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = -.225;
			offset_y = -.1125;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = .3375;
			offset_y = -.225;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = -.3375;
			offset_y = -.3375;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);

			offset_x = -.1125;
			offset_y = -.45;
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x + offset_x, y + OpenGL_offset_y + offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth, samplesPerPixel);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);
			break;
		default:
			// one sample, no multisampling
			start = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x, y + OpenGL_offset_y, 0, 1) * getRenderMatrixInverse());
			end = osg::Vec4dH(osg::Vec4d(x + OpenGL_offset_x, y + OpenGL_offset_y, 1, 1) * getRenderMatrixInverse());
			ray = new PrimaryRay(start.project(), end.project(), x, y, maxRecursionDepth);
			if (debugInfo) {
				ray->displayDebugInfo();
			}
			tracer->addRay(ray);
			ray->setOptions(_referenceOptions);
			break;
	}
}


const osg::Matrixd & RayManager::getRenderMatrixInverse() const {
	return *(_renderMatrixInverse.get());
}


void RayManager::tracePoint(const uint x, const uint y) {
	updateRenderMatrixInverse();

	//	//output master and render cameras
	//	osgDB::writeNodeFile(*(_masterCamera.get()), "masterCamera.osg");
	//	osgDB::writeNodeFile(*(_renderCamera.get()), "renderCamera.osg");


	//	osg::Vec4dH start, end;
	//	start = osg::Vec4dH(osg::Vec4d(x + 0.5, y + 0.5, 0, 1) * getPrimaryRayTransformMatrixInverse());
	//	end = osg::Vec4dH(osg::Vec4d(x + 0.5, y + 0.5, 1, 1) * getPrimaryRayTransformMatrixInverse());
	//
	//	std::cout << start.x() << " " << start.y() << " " << start.z() << " - " << end.x() << " " << end.y() << " " << end.z() << std::endl;
	//
	//	osg::Vec3d s, e;
	//	s = start.project();
	//	e = end.project();
	//
	//	std::cout << s.x() << " " << s.y() << " " << s.z() << " - " << e.x() << " " << e.y() << " " << e.z() << std::endl;
	//	s = e - s;
	//	s.normalize();
	//	std::cout << s.x() << " " << s.y() << " " << s.z() << std::endl;



	//	PrimaryRay* ray = new PrimaryRay(x + 0.5, y + 0.5, x, y, MAX_RECUSION_DEPTH);


	_pointTracer->reset();
	createAndAssignBeam(_pointTracer, x, y, true);
	//	_pointTracer->addRay(ray);

	//	s = ray->getStartWorldPoint();
	//	e = ray->getEndWorldPoint();
	//	std::cout << s.x() << " " << s.y() << " " << s.z() << " - " << e.x() << " " << e.y() << " " << e.z() << std::endl << std::endl;


	_pointTracer->trace();
	osg::Image *renderImage;
	renderImage = _outputBuffer;
	_pointTracer->writeImage(renderImage);
	renderImage->dirty();
	// request the ray display to redraw
	if (_renderView.get()) {
		_renderView->requestRedraw();
	}

	//	std::cout << ".";
}


void RayManager::prepareOutputImage() {
	osg::Image *renderImage;
	renderImage = _outputBuffer;
	const uint width = renderImage->s();
	const uint height = renderImage->t();

#ifndef NOSCENECOPY
	osg::Image *inputImage = _sceneSnapshot;
	// copy the input picture to the output picture
	this->inputSnapshot(renderImage, inputImage);
#else
	// vymazanie obsahu obrazka
	for (uint x = 0; x < width; x += 1) {
		for (uint y = 0; y < height; y += 1) {
			RayManager::setImagePointColorFtoD(renderImage, osg::Vec2(x, y), osg::Vec4d(0, 0, 0, 1));
		}
	}
	renderImage->dirty();
#endif
}


int RayManager::cancel() {
	if (isActive()) {
		if (_distribution == RayManager::DYNAMIC) {
			deactivate();
		}
		for (uint i = 0; i < _tracerThreads.size(); i += 1) {
			_tracerThreads[i]->cancel();
		}
		// wait for all the tracer threads to finish
		uint countFinished;
		do {
			countFinished = 0;
			OpenThreads::Thread::microSleep(200000); // 5 times a second is enough
			for (uint i = 0; i < _tracerThreads.size(); i += 1) {
				countFinished += _tracerThreads[i]->isRunning() ? 0 : 1;
			}
		} while (countFinished < _tracerThreads.size() && isActive());
		deactivate();
	}
	// then wait for the the thread to stop running.
	while (isRunning()) {
		OpenThreads::Thread::YieldCurrentThread();
	}
	return 0;
}


void RayManager::startPointTracing() {
	_pointTracing = true;
	//	osg::Image *renderImage;
	//	renderImage = _outputBuffer;
	//	uint width = renderImage->s();
	//	uint height = renderImage->t();
	// vymazanie obsahu obrazka
	//	for (uint x = 0; x < width; x += 1) {
	//		for (uint y = 0; y < height; y += 1) {
	//			RayManager::setImagePointColorFtoD(renderImage, osg::Vec2(x, y), osg::Vec4d(0, 0, 0, 1));
	//		}
	//	}
	//	renderImage->dirty();

	// collect light sources in the scene
	WorldLightSourceContainer * lc = new WorldLightSourceContainer();
	collectLightSources(*lc);

	osg::Node * scene = _scene.get();

	if (!_pointTracer) {
		_pointTracer = new Tracer(scene, lc);
	} else {
		_pointTracer->reset();
		_pointTracer->setScene(scene);
		_pointTracer->setWorldLightSourceContainer(lc);
	}

}


void RayManager::stopPointTracing() {
	_pointTracing = false;
}


void RayManager::collectLightSources(WorldLightSourceContainer & lc) const {
	osg::ref_ptr<LightSourceVisitor> lVisitor = new LightSourceVisitor(lc.sources);
	_scene->accept(*(lVisitor.get()));

	for (LightSources::iterator lsIt = lc.sources.begin(); lsIt != lc.sources.end(); lsIt++) {

		osg::MatrixList worldMatrices = (*lsIt)->getWorldMatrices();

		// take the first world matrix
		osg::Matrix m = worldMatrices.front();
		//		osg::Matrix m = worldMatrices.back();

		lc.positions.push_back((*lsIt)->getLight()->getPosition() * m);
		osg::Vec3d direction = (*lsIt)->getLight()->getDirection() * m;
		direction.normalize();
		lc.directions.push_back(direction);

#ifdef PRINT_LIGHT_SOURCES
		osg::Light *light = (*lsIt)->getLight();
		osg::Vec4d v = light->getDiffuse();
		std::cout << "type: ";
		if (!light->getPosition().w()/* == 0.0f*/) {
			// directional light - sun
			std::cout << "directional";
		} else if (light->getSpotCutoff() == 180.0f) {
			// point light - a bulb
			std::cout << "point";
		} else {
			// spot light - a table lamp


			std::cout << "spot";
		}
		std::cout << std::endl;
		osg::Vec4d pos = lc.positions.back();
		std::cout << "position " << pos.x() << " " << pos.y() << " " << pos.z() << " " << pos.w() << std::endl;
		std::cout << "diff " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << std::endl;
		v = light->getSpecular();
		std::cout << "spec " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << std::endl;
		std::cout << "Paths to light: " << worldMatrices.size() << std::endl;
#endif
	}
}


void RayManager::inputSnapshot(osg::Image *image, const osg::Image * source) {
	image->copySubImage(0, 0, 0, source);
}


void RayManager::setImagePointColor(osg::Image *img, const osg::Vec2 pointCoords, const osg::Vec4 color) {
	uchar *point = img->data(pointCoords.x(), pointCoords.y());
	point[0] = color.r();
	point[1] = color.g();
	point[2] = color.b();
	point[3] = color.a();
}


void RayManager::setImagePointColorFtoD(osg::Image *img, const osg::Vec2 pointCoords, const osg::Vec4 color) {
	uchar *point = img->data(pointCoords.x(), pointCoords.y());
	point[0] = (int) (255 * color.r());
	point[1] = (int) (255 * color.g());
	point[2] = (int) (255 * color.b());
	point[3] = (int) (255 * color.a());
}


bool RayManager::isPointTracing() const {
	return (bool) _pointTracing;
}


void RayManager::setProgress(const uchar progress) {
	if (progress > 100) {
		_progress.exchange(100);
	} else {
		_progress.exchange(progress);
	}
}


uchar RayManager::getProgress() const {
	return _progress;
}


bool RayManager::isReady() const {
	return (_progress == 0 || _progress == 100) && !this->isActive();
}


void RayManager::activate() {
	_active.exchange(1);
}


void RayManager::deactivate() {
	_active.exchange(0);
}


bool RayManager::isActive() const {
	return (bool) _active;
}


bool RayManager::hasDedicatedThreads() const {
	return (bool) (_threadsPerProcessor >= 0 || _threadCount > 0);
}


osg::Timer_t RayManager::getRuntime() const {
	return _lastRuntime;
}


RayManager* RayManager::setRecursionDepth(uchar depth) {
	_recursionDepth = depth;
	return this;
}


uchar RayManager::getRecursionDepth() const {
	return _recursionDepth;
}


RayManager* RayManager::setThreadCount(ushort x) {
	_threadCount = x;
	return this;
}


ushort RayManager::getThreadCount() const {
	return _threadCount;
}


RayManager* RayManager::enableTextures(bool x) {
	_referenceOptions.textures = x;
	return this;
}


RayManager* RayManager::enableBilinearFiltering(bool x) {
	_referenceOptions.bilinearTextureFiltering = x;
	return this;
}


RayManager* RayManager::enableRecursion(bool x) {
	_referenceOptions.recursion = x;
	return this;
}


RayManager* RayManager::enableReflections(bool x) {
	_referenceOptions.reflections = x;
	return this;
}


RayManager* RayManager::enableRefractions(bool x) {
	_referenceOptions.refractions = x;
	return this;
}


RayManager* RayManager::enablePhongShading(bool x) {
	_referenceOptions.phongShading = x;
	return this;
}


RayManager* RayManager::enableSelfShadows(bool x) {
	_referenceOptions.selfShadows = x;
	return this;
}


RayManager* RayManager::setTrimmer(uchar trimmer, double value) {
	if (trimmer == 0 || trimmer == 1) {
		_referenceOptions.reflectionTrimmer = value;
	}
	if (trimmer == 0 || trimmer == 2) {
		_referenceOptions.refractionTrimmer = value;
	}
	return this;
}


RayManager* RayManager::setSamplesPerPixel(uchar x) {
	_samplesPerPixel = x;
	return this;
}


RayManager* RayManager::setChunkSize(uint x) {
	_chunkSize = x;
	return this;
}


void RayManager::writeOutput(const std::string outputPath, uint timestamp) const {
	if (this->isReady()) {
		osg::Image *i = (_outputBuffer.get());
		std::stringstream ss;
		uint t = timestamp;
		if (!timestamp) {
			t = time(0);
		}
		ss << outputPath << "rayme_" << t << "_" << i->s() << "x" << i->t() << "_" << getRuntime() << "s" << ".bmp";
		if (writeImage(ss.str(), *i)) {
			std::cout << "screens saved as " << ss.str() << std::endl;
		}
	}
}


void RayManager::writeSnapshot(const std::string outputPath, uint timestamp) const {
	std::stringstream ss;
	uint t = timestamp;
	if (!timestamp) {
		t = time(0);
	}
	ss << outputPath << "rayme_" << t << "_scene" << ".bmp";
	writeImage(ss.str(), *(_sceneSnapshot.get()));
}


bool RayManager::writeImage(const std::string filename, const osg::Image& image) const {
	return osgDB::writeImageFile(image, filename);
}


void RayManager::storeMasterCamera(const std::string outputPath, uint timestamp) const {
	std::stringstream ss;
	uint t = timestamp;
	if (!timestamp) {
		t = time(0);
	}
	ss << outputPath << "rayme_" << t << "_camera" << ".osg";
	osg::Camera* cam = new osg::Camera();
	cam->setViewport(_masterCamera->getViewport());
	cam->setProjectionMatrix(_masterCamera->getProjectionMatrix());
	cam->setViewMatrix(_masterCamera->getViewMatrix());
	osgDB::writeNodeFile(*cam, ss.str());
}


void RayManager::setConsoleProgressPrintoutThreshold(char x) {
	_printProgressToConsoleThreshold = x;
}

