/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#ifndef RAYME_RAYMANAGER_H
#define	RAYME_RAYMANAGER_H


#include <OpenThreads/Atomic>
#include <OpenThreads/Thread>
#include <osgViewer/Viewer>
#include <osg/Matrixd>

#include "common.h"
#include "TracerThread.h"

namespace Rayme {

	//typedef std::function<void> RefreshCbFunc;
	class RayManager : public OpenThreads::Thread {
	public:
		RayManager(osg::Camera* camera, osg::Node* scene, osg::Image* outputBuffer, osg::Image* sceneSnapshot = NULL, osgViewer::View* renderView = NULL);
		virtual ~RayManager();
		enum RayDistributionModel {
			STATIC = 0, DYNAMIC = 1
		};


	public:
		void requestNewFrame();
		void startPointTracing();
		void stopPointTracing();
		bool isPointTracing() const;
		void tracePoint(const uint x, const uint y);
		bool isReady() const;
		static void setImagePointColor(osg::Image *img, const osg::Vec2 pointCoords, const osg::Vec4 color);
		static void setImagePointColorFtoD(osg::Image *img, const osg::Vec2 pointCoords, const osg::Vec4 color);
		virtual int cancel();
		virtual void run();
		bool isActive() const;
		bool hasDedicatedThreads() const;
		osg::Timer_t getRuntime() const;
		void writeOutput(const std::string outputPath, uint timestamp = 0) const;
		void writeSnapshot(const std::string outputPath, uint timestamp = 0) const;
		void storeMasterCamera(const std::string outputPath, uint timestamp = 0) const;
		void setRenderMatrixInverse(osg::Matrixd& matrix);
		void setConsoleProgressPrintoutThreshold(char x = -1);


	public: // UI port methods
		uchar getProgress() const;
		RayManager* setRecursionDepth(uchar depth);
		uchar getRecursionDepth() const;
		RayManager* setThreadCount(ushort x);
		ushort getThreadCount() const;
		RayManager* enableTextures(bool x = true);
		RayManager* enableBilinearFiltering(bool x = true);
		RayManager* enablePhongShading(bool x = true);
		RayManager* enableRecursion(bool x = true);
		RayManager* enableReflections(bool x = true);
		RayManager* enableRefractions(bool x = true);
		RayManager* enableSelfShadows(bool x = true);
		RayManager* setSamplesPerPixel(uchar x);
		RayManager* setChunkSize(uint x);
		RayManager* setTrimmer(uchar trimmer, double value);


	protected:
		void initThreads(const uint threadCount);
		void inputSnapshot(osg::Image *image, const osg::Image *source);
		void prepareOutputImage();
		void setProgress(const uchar progress);
		void activate();
		void deactivate();
		void collectLightSources(WorldLightSourceContainer &lc) const;
		void startSceneRayTracing(RayDistributionModel distribution, const uint threadCount);
		void createAndAssignBeam(Tracer* tracer, const ushort x, const ushort y, bool debugInfo = false);
		void updateRenderMatrixInverse();
		const osg::Matrixd& getRenderMatrixInverse() const;
		bool writeImage(const std::string filename, const osg::Image& image) const;


	protected:
		OpenThreads::Atomic _active;
		OpenThreads::Atomic _progress;
		osg::ref_ptr<osg::Node> _scene;
		osg::ref_ptr<osg::Camera> _masterCamera;
		osg::ref_ptr<osg::RefMatrixd> _renderMatrixInverse;
		osg::ref_ptr<osgViewer::View> _renderView;
		osg::ref_ptr<osg::Image> _outputBuffer;
		osg::ref_ptr<osg::Image> _sceneSnapshot;
		RayDistributionModel _distribution;
		uchar _samplesPerPixel;
		uchar _recursionDepth;
		uint _processorCount;
		/**
		 * Used for overriding the automatic thread count settings.
		 */
		ushort _threadCount;
		char _threadsPerProcessor;
		uint _chunkSize;
		std::vector<TracerThread *> _tracerThreads;
		bool _pointTracing;
		Tracer* _pointTracer;
		osg::Timer_t _lastRuntime;
		RayIntersector::Options _referenceOptions;
		char _printProgressToConsoleThreshold;

	};
}

#endif	/* RAYME_RAYMANAGER_H */

