/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osg/StateSet>
#include <osg/StateAttribute>
#include <osg/Material>
#include <osg/Geometry>
#include <osg/Vec3d>
#include "RefractedRay.h"
#include "PhotorealismData.h"
#include "ReflectedRay.h"

#define PHONG_SHADING
#define INTERNAL_REFLECTION //TODO

using namespace Rayme;


RefractedRay::RefractedRay(double refractiveIndex, osg::Vec3d start, osg::Vec3d direction, uchar currentDepth, uchar maxDepth) :
SecondaryRay(start + direction * POINT_PRECISION, start + direction * POINT_INFINITY, 1.0, currentDepth, maxDepth), _n1(refractiveIndex), _internalDepth(0) {
}


RefractedRay::RefractedRay(double refractiveIndex, osg::Vec3d start, osg::Vec3d direction, double weight, uchar currentDepth, uchar maxDepth) :
SecondaryRay(start + direction * POINT_PRECISION, start + direction * POINT_INFINITY, weight, currentDepth, maxDepth), _n1(refractiveIndex), _internalDepth(0) {
}


RefractedRay::RefractedRay(double refractiveIndex, osg::Vec3d start, osg::Vec3d direction, double weight, uchar currentDepth, uchar maxDepth, uchar currentInternalDepth) :
SecondaryRay(start + direction * POINT_PRECISION, start + direction * POINT_INFINITY, weight, currentDepth, maxDepth), _n1(refractiveIndex), _internalDepth(currentInternalDepth) {
}


RefractedRay::~RefractedRay() {
}


RefractedRay* RefractedRay::evaluate() {
	if (_evaluated || !_tracer) {
		return this;
	}
	if (_weight < LIGHT_INTENSITY_CUTOFF) {
		_color.set(0, 0, 0, 1);
		return this;
	}

	if (!_traced) {
		_tracer->traceSingle(this);
	}

	if (_debugInfo) {
		std::cout << "(" << (int) _currentDepth << "/" << (int) _internalDepth << ") " << std::endl;
		//		std::cout << _start.x() << " " << _start.y() << " " << _start.z() << " -> ";
		//		std::cout << _end.x() << " " << _end.y() << " " << _end.z() << std::endl;
	}

	if (containsIntersections() && hasValidIntersections()) {

		/**
		 * The intersection denoting the point hit by the ray.
		 */
		const osgUtil::LineSegmentIntersector::Intersection& intersection = getValidIntersection();
		/**
		 * The Drawable hit by this ray.
		 */
		osg::Drawable *collisionDrawable = intersection.drawable.get();
		/**
		 * The point of collision.
		 */
		const osg::Vec3d collisionPoint = intersection.getWorldIntersectPoint();
		/**
		 * Normal to the surface at the collision point, pointing out of the object.
		 * If Phong shading enabled, the value is interpolated later.
		 */
		osg::Vec3d outerSurfaceNormal = intersection.getWorldIntersectNormal();
		/**
		 * Normal to the surface at the collision point, pointing into the object.
		 * The inverse of outerSurfaceNormal.
		 */
		osg::Vec3d innerInterfaceNormal = intersection.getWorldIntersectNormal();

		/**
		 * The direction of the ray. Normalized.
		 */
		osg::Vec3d rayDirectionVector = getDirection();
		rayDirectionVector.normalize();

		if (_debugInfo) {
			std::cout << "* " << _start.x() << " " << _start.y() << " " << _start.z();
			std::cout << " T " << collisionPoint.x() << " " << collisionPoint.y() << " " << collisionPoint.z() << std::endl;
			std::cout << "-> " << rayDirectionVector.x() << " " << rayDirectionVector.y() << " " << rayDirectionVector.z() << std::endl;
		}



		/**
		 * The Geometry hit. Cast from the Drawable.
		 */
		osg::Geometry *collisionGeometry = collisionDrawable ? collisionDrawable->asGeometry() : NULL;
		if (collisionGeometry) {
#ifdef PHONG_SHADING
			if (_options.phongShading) {
				// check the normal binding and if per-vertex normals are used, interpolate the vertex normals (phong interpolation)
				if (collisionGeometry->getNormalBinding() == osg::Geometry::BIND_PER_VERTEX) {
					const osg::Array *a = collisionGeometry->getNormalArray();
					if (a->getType() == osg::Array::Vec3ArrayType) {
						const osg::Vec3Array *drawableNormalArray = dynamic_cast<const osg::Vec3Array*> (a);
						if (drawableNormalArray) {
							osg::Vec3 interpolatedNormal(0, 0, 0);
							osg::Matrix *matrix = intersection.matrix.get();
							osg::Matrix inverse;
							inverse.invert(*matrix);
							for (uint i = 0; i < intersection.indexList.size(); ++i) {
								osg::Vec3d tmp =
									  (*drawableNormalArray)[intersection.indexList[i]] * intersection.ratioList[i];
								tmp = matrix->valid() ? osg::Matrix::transform3x3(inverse, tmp) : tmp;
								interpolatedNormal += tmp;
							}
							outerSurfaceNormal = interpolatedNormal;
						}//if drawable normal
					}//if array of Vec3 points
				}//if per vertex normal binding
			}
#endif // phong shading
		}

		// normalize normal
		outerSurfaceNormal.normalize();
		innerInterfaceNormal = -outerSurfaceNormal;

		// add the light contribution from recursive secondary (reflection and refraction) rays
		const double reflectance = computeReflectanceSchlickApprox(rayDirectionVector, innerInterfaceNormal, _n1, 1.0);
		const double refractance = 1 - reflectance;
		osg::Vec4d color(0, 0, 0, 0);

		if (_debugInfo) {
			std::cout << "reflectance: " << reflectance << std::endl;
		}

		if (refractance * _weight >= LIGHT_INTENSITY_CUTOFF) {
			if (_debugInfo) {
				std::cout << "dot. " << (rayDirectionVector * innerInterfaceNormal) << std::endl;
			}

			osg::Vec3d refractionVector = computeRefractionVector(rayDirectionVector, innerInterfaceNormal, _n1, 1.0);

			//			if (_debugInfo) {
			//				std::cout << "norm " << innerInterfaceNormal.x() << " " << innerInterfaceNormal.y() << " " << innerInterfaceNormal.z() << std::endl;
			//				std::cout << "refr " << refractionVector.x() << " " << refractionVector.y() << " " << refractionVector.z() << " (" << _n1 << ")" << std::endl;
			//			}

			if (refractionVector.length2() > POINT_PRECISION) {
				osg::ref_ptr<ReflectedRay> secondaryRay = new ReflectedRay(collisionPoint, refractionVector, refractance * _weight, _currentDepth, _maxDepth); //NOTE: note the current depth not being incremented
				secondaryRay->setTracer(_tracer);
				secondaryRay->setOptions(_options);
				secondaryRay->displayDebugInfo(_debugInfo);
				secondaryRay->evaluate();
				color += secondaryRay->getWeightedColor();

			} else {
				// TIR (total inner reflection)
			}
		}

#ifdef INTERNAL_REFLECTION
		if (reflectance * _weight >= LIGHT_INTENSITY_CUTOFF) {
			osg::Vec3d reflectionVector = computeReflectionVector(rayDirectionVector, innerInterfaceNormal);
			if (true /* reflectance high enough*/ && _internalDepth < _maxDepth) {
				osg::ref_ptr<RefractedRay> secondaryRay = new RefractedRay(_n1, collisionPoint, reflectionVector, reflectance * _weight, _currentDepth, _maxDepth, _internalDepth + 1); //NOTE: note the INTERNAL depth being incremented
				secondaryRay->setTracer(_tracer);
				secondaryRay->setOptions(_options);
				secondaryRay->displayDebugInfo(_debugInfo);
				secondaryRay->evaluate();
				color += secondaryRay->getWeightedColor();
			}
		}
#endif

		_color.set(CLAMP(color.r()), CLAMP(color.g()), CLAMP(color.b()), 1.0);
		_valid = true;
	} else {
		if (_debugInfo) {
			std::cout << "refracted missed! bug? ";
			std::cout << "rec.depth: " << (int) _currentDepth << ", s-e: ";
			std::cout << getStart().x() << " " << getStart().y() << " " << getStart().z() << " - " << getEnd().x() << " " << getEnd().y() << " " << getEnd().z() << std::endl;
		}
	}

	_evaluated = true;
	return this;
}


void RefractedRay::selfTrace(Tracer const& tracer) {
	if (_debugInfo) {
		//		std::cout << "self tracing REFRACTED" << std::endl;
	}

	osgUtil::IntersectionVisitor intersectionVisitor(this);
	intersectionVisitor.setUseKdTreeWhenAvailable(true);
	tracer.getScene()->accept(intersectionVisitor);
	setTraced();
}

