/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osgUtil/LineSegmentIntersector>
#include <osg/Drawable>
#include <osg/StateSet>
#include <osg/Material>

#include "ShadowRay.h"
#include "PhotorealismData.h"

using namespace Rayme;


ShadowRay::ShadowRay(osg::Node* scene, const osgUtil::LineSegmentIntersector::Intersection& intersection, const osg::Vec3d lightPosition, bool selfShadow) {
	_fullShadow = false;
	_direct = true;
	_lightContribution = 1.0;

	osg::ref_ptr<osgUtil::LineSegmentIntersector> shadowRay =
		  new osgUtil::LineSegmentIntersector(intersection.getWorldIntersectPoint()
		  //+(intersection.getWorldIntersectPoint() - lightPosition)*0.0001 // this was used to offset the shadow ray from the starting point, but is faulty since rev. 61
		  , lightPosition);

	osgUtil::IntersectionVisitor intersectionVisitor(shadowRay.get());
	scene->accept(intersectionVisitor);

	// now check whether the obscuring objects (at intersections) are not parts of the light geometry and not the tracing ray collision point itself
	if (shadowRay->containsIntersections()) {
		bool transparentThrough = false;
		osgUtil::LineSegmentIntersector::Intersections shadowHits = shadowRay->getIntersections();
		uint shadowRayHitCount = shadowHits.size();

		bool inside = false;
		for (osgUtil::LineSegmentIntersector::Intersections::iterator it = shadowHits.begin(); it != shadowHits.end(); ++it) {
			double transparency = 0.0;

			osg::Vec3d start = shadowRay->getStart().getVectorPart(); // no need to project
			osg::Vec3d ip = (*it).getWorldIntersectPoint();
			{
				osg::Vec3d surfaceNormal = (*it).getWorldIntersectNormal();
				surfaceNormal.normalize();
				osg::Vec3d direction = shadowRay->getEnd().getVectorPart() - start;
				direction.normalize();
				double checkAngle = surfaceNormal * direction;
				if (checkAngle < 0) {
					inside = true;
					//					std::cout << "into object (" << checkAngle << ")"<<std::endl;
				} else {
					inside = false;
					//					std::cout << "out of obj. (" << checkAngle << ")"<<std::endl;
				}
			}


			osg::Node *node = !(*it).nodePath.empty() ? (*it).nodePath.back() : NULL;
			if ((*it).drawable.get() == intersection.drawable.get()) {

				// if the occluding point is equal to the starting point - ignore
				if (selfShadow) {
					if ((start - ip).length2() < POINT_PRECISION) {
						shadowRayHitCount -= 1;
						//					std::cout << "self"<<std::endl;
					} else {
						// do the transparency check
						transparency = transparencyCheck(*it);
						if (transparency > 0) {
							if (inside) { // entering a transparent object
								_lightContribution *= transparency;
							} else { // leaving a transparent object
								if (it != shadowHits.begin()) {
									// let's make the thickness of the object count
									const double lengthFactor = (ip - (*(--it)).getWorldIntersectPoint()).length2() * 0.2;
									++it;
									_lightContribution = _lightContribution / (lengthFactor + _lightContribution);
								} else {
									// a bug? this shouldn't ever happen
								}
							}


							shadowRayHitCount -= 1;
							transparentThrough = true;
							inside = inside ? false : true;
						} else {
							break;
						}
					}
				} else {
					// if no self-shadow can be cast, a drawable cannot occlude itself
					shadowRayHitCount -= 1;
				}

			} else if (node) {

				// the occluding object is part of the light's geometry (e.g. a bulb) - ignore
				PhotorealismData photorealismData(node);
				if (photorealismData.contains("Material.castShadow")) {
					std::stringstream in(photorealismData.get("Material.castShadow"));
					bool castShadow;
					in >> castShadow;
					if (!castShadow) {
						// the node hit is part of the light geometry - no shadow is cast
						shadowRayHitCount -= 1;
					}
				} else {
					// do the transparency check
					transparency = transparencyCheck(*it);
					if (transparency > 0) {
						if (inside) { // entering a transparent object
							_lightContribution *= transparency;
						} else { // leaving a transparent object
							if (it != shadowHits.begin()) {
								// let's make the thickness of the object count
								const double lengthFactor = (ip - (*(--it)).getWorldIntersectPoint()).length2() * 0.5;
								++it;
								_lightContribution = _lightContribution / (lengthFactor + _lightContribution);
							} else {
								// a bug? this shouldn't ever happen
							}
						}

						shadowRayHitCount -= 1;
						transparentThrough = true;
						inside = inside ? false : true;
					} else {
						break;
					}
				}

			} else {
				// should this never happen?
				break;
			}
		}

		if (shadowRayHitCount > 0) {
			// the light is obscured
			_fullShadow = true;
		} else {
			if (transparentThrough) {
				_direct = false;
			}
		}
	}
}


ShadowRay::~ShadowRay() {
}


double ShadowRay::transparencyCheck(osgUtil::LineSegmentIntersector::Intersection const& intersection) const {
	double transparency = 0.0;

	osg::Drawable *collisionDrawable = intersection.drawable.get();
	osg::StateSet *collisionStateSet = collisionDrawable ? collisionDrawable->getStateSet() : NULL;
	osg::Material *material = collisionStateSet ? dynamic_cast<osg::Material*> (collisionStateSet->getAttribute(osg::StateAttribute::MATERIAL)) : NULL;
	if (material) {
		osg::Vec4d tmp = material->getDiffuse(osg::Material::FRONT);
		transparency = 1 - tmp.a();
	}

	return CLAMP(NONNEGATIVE(transparency));
}


bool ShadowRay::castsFullShadow() const {
	return _fullShadow;
}


double ShadowRay::getLightContribution() const {
	return _fullShadow ? 0.0 : (_direct ? 1.0 : CLAMP(_lightContribution));
}
