/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osg/Camera>
#include <osgUtil/IntersectionVisitor>
#include <osg/StateSet>
#include <osg/StateAttribute>
#include <osg/Material>

#include "Tracer.h"
#include "PrimaryRay.h"

//#define MEASURE_TIME 

using namespace Rayme;


Tracer::Tracer(osg::Node* scene, WorldLightSourceContainer* lc, osgUtil::IntersectorGroup* rays) :
_scene(scene), _lightSourceContainer(lc), _rays(rays) {
	init();
}


Tracer::Tracer(osg::Node* scene, WorldLightSourceContainer* lc) :
_scene(scene), _lightSourceContainer(lc) {
	init();
}


void Tracer::init() {
	_rays = new osgUtil::IntersectorGroup();
	_eye_set = false;
}


Tracer::~Tracer() {
	delete(_rays);
	//TODO toto opravit ! (nikdy sa nezavola ?)
	//std::cout << "tracer destroyed";
}


void Tracer::trace() {
#ifdef MEASURE_TIME
	osg::Timer_t frameStartTick = osg::Timer::instance()->tick();
#endif

	// intersection visitor with the use-k-D-trees hint
	osgUtil::IntersectionVisitor intersectionVisitor(_rays.get());
	intersectionVisitor.setUseKdTreeWhenAvailable(true);

	// cast the primary rays
	_scene->accept(intersectionVisitor);

	for (osgUtil::IntersectorGroup::Intersectors::iterator intersector_itr = _rays->getIntersectors().begin(); intersector_itr != _rays->getIntersectors().end(); ++intersector_itr) {
		RayIntersector* ray = dynamic_cast<RayIntersector*> ((*intersector_itr).get());
		if (ray) {
			ray->setTraced();
		}
	}//foreach intersector

	// get the eye position for whole tracer
	//	setEye();

	// are there any intersections?
	if (_rays->containsIntersections()) {

		// enter the evaluation loop - this is provided for easier threading implementation
		evalLoop();

	}

#ifdef MEASURE_TIME
	osg::Timer_t frameEndTick = osg::Timer::instance()->tick();
	_totaltime = osg::Timer::instance()->delta_s(frameStartTick, frameEndTick);
#endif
}


void Tracer::traceSingle(RayIntersector* ray) {
	ray->selfTrace(*this);
}


void Tracer::evalLoop() {
	// cycle through the intersectors and respective intersections
	osgUtil::IntersectorGroup::Intersectors& intersectors = _rays->getIntersectors();
	for (osgUtil::IntersectorGroup::Intersectors::iterator intersector_itr = intersectors.begin(); intersector_itr != intersectors.end(); ++intersector_itr) {

		RayIntersector* ray = dynamic_cast<RayIntersector*> ((*intersector_itr).get());

		if (ray) {
			ray->evaluate();
		}

		++_finishedRayCount;

	}//foreach intersector
}


//void Tracer::setEye() {
//	if (!_eye_set) {
//		osg::Camera * cam = dynamic_cast<osg::Camera*> (getScene());
//		if (cam) {
//			_eye = osg::Vec3d(0.0, 0.0, 0.0) * osg::Matrixd::inverse(cam->getViewMatrix());
//			_eye_set = true;
//		} else {
//			std::cout << "Error: No camera to determine the eye position." << std::endl;
//		}
//	}
//}


osgUtil::IntersectorGroup * Tracer::getRays() const {
	return _rays.get();
}


void Tracer::setScene(osg::Node * scene) {
	_scene = scene;
}


osg::Node * Tracer::getScene() const {
	return _scene;
}


void Tracer::setWorldLightSourceContainer(WorldLightSourceContainer * lc) {
	_lightSourceContainer = lc;
}


WorldLightSourceContainer * Tracer::getWorldLightSourceContainer() const {
	return _lightSourceContainer;
}


void Tracer::addRay(RayIntersector * ray) {
	ray->setTracer(this);
	osgUtil::Intersector * i = dynamic_cast<osgUtil::Intersector*> (ray);
	if (i) {
		this->getRays()->addIntersector(i);
	}
}


void Tracer::setRays(osgUtil::IntersectorGroup * rays) {
	_rays = rays;
}


void Tracer::reset() {
	_rays = new osgUtil::IntersectorGroup;
	_totaltime = 0;
	_eye_set = false;
}


void Tracer::writeImage(osg::Image * img) const {
	osgUtil::IntersectorGroup::Intersectors& intersectors = _rays->getIntersectors();
	for (osgUtil::IntersectorGroup::Intersectors::iterator intersector_itr = intersectors.begin(); intersector_itr != intersectors.end(); ++intersector_itr) {
		PrimaryRay* ray = dynamic_cast<PrimaryRay*> ((*intersector_itr).get());
		if (ray) {
			if (ray->isMultiSample()) {
				addImagePointColorFtoD(img, ray->x(), ray->y(), ray->getColor() * (1.0 / ray->samplesPerPixel()));
			} else {
				setImagePointColorFtoD(img, ray->x(), ray->y(), ray->getColor());
			}
		}
	}
}


void Tracer::setImagePointColor(osg::Image *img, const uint x, const uint y, const osg::Vec4 color) const {
	uchar *point = img->data(x, y);
	point[0] = color.r();
	point[1] = color.g();
	point[2] = color.b();
	point[3] = color.a();
}


void Tracer::setImagePointColorFtoD(osg::Image *img, const uint x, const uint y, const osg::Vec4 color) const {
	uchar *point = img->data(x, y);
	point[0] = (int) (255 * color.r());
	point[1] = (int) (255 * color.g());
	point[2] = (int) (255 * color.b());
	point[3] = (int) (255 * color.a());
}


void Tracer::addImagePointColorFtoD(osg::Image *img, const uint x, const uint y, const osg::Vec4 color) const {
	uchar *point = img->data(x, y);
	double tmp;
	point[0] = (int) ((tmp = 255 * color.r() + point[0]) > 255 ? 255 : tmp);
	point[1] = (int) ((tmp = 255 * color.g() + point[1]) > 255 ? 255 : tmp);
	point[2] = (int) ((tmp = 255 * color.b() + point[2]) > 255 ? 255 : tmp);
	point[3] = (int) ((tmp = 255 * color.a() + point[3]) > 255 ? 255 : tmp);
}


osg::Timer_t Tracer::getTotalTime() const {
#ifndef MEASURE_TIME
	std::cout << "WARNING: Tracer time measuring not supported!" << std::endl;
#endif
	return _totaltime;
}


uchar Tracer::getProgress() const {
	return _finishedRayCount * 100 / getRayCount();
}


uint Tracer::getFinishedCount() const {
	return _finishedRayCount;
}


uint Tracer::getRayCount() const {
	return _rays->getIntersectors().size();
}


osg::Vec3d Tracer::getEye() const {
	return _eye;
}
