/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


#include <osgDB/ReadFile>
#include <osgDB/WriteFile>

#include "UILayer.h"

using namespace Rayme;


UILayer::UILayer() :
_arguments(NULL), _rm(NULL), _wm(NULL) {
	setDefaults();
}


UILayer::UILayer(int& argc, char** argv) :
_arguments(NULL), _rm(NULL), _wm(NULL) {
	setDefaults();
	parseArguments(argc, argv);
}


UILayer::UILayer(int& argc, char** argv, RayManager* rm, WindowManager* wm) :
_arguments(NULL), _rm(rm), _wm(wm) {
	//	bindWM();
	setDefaults();
	parseArguments(argc, argv);
}


UILayer::UILayer(RayManager* rm, WindowManager* wm) :
_arguments(NULL), _rm(rm), _wm(wm) {
	//	bindWM();
	setDefaults();
}


UILayer::~UILayer() {
	//		std::cout <<"~ui"<<std::endl;
	delete(_arguments);
}


const UIState& UILayer::getUIState() const {
	return _state;
}


void UILayer::parseArguments(int& argc, char** argv) {
	if (_arguments != NULL) {
		delete(_arguments);
	}
	_arguments = new osg::ArgumentParser(&argc, argv);
	_error = 0;

	// use an ArgumentParser object to manage the program arguments
	osg::ArgumentParser* argumentParser = _arguments;
	osg::ApplicationUsage &au = *argumentParser->getApplicationUsage();

	// application usage
	au.setApplicationName(argumentParser->getApplicationName());
	au.setDescription(argumentParser->getApplicationName() +
		  " is a 3D model viewer with the ability to render realistic screenshots using ray-tracing method.");
	au.setCommandLineUsage(argumentParser->getApplicationName() + " <sceneFileName> [options] ");

	au.addCommandLineOption("[-r|--resolution] <width> <height>", "Set the initial resolution of the viewers and output image.");
	au.addCommandLineOption("[-o|--output-path] <path>", "Set the path for possible output files - screenshots, models etc.");
	au.addCommandLineOption("--console", "Run the application in console mode without windows. The output is saved automatically.");
	au.addCommandLineOption("--camera <cameraFile>", "Set the initial camera. For console mode.");
	au.addCommandLineOption("--samples <number>", "Set the number of samples per pixel. Accepted values: 1, 4");
	au.addCommandLineOption("--reflection-trimmer <number>", "Set the reflections trimmer. Reduces reflections. Accepted values: 0.0-1.0");
	au.addCommandLineOption("--refraction-trimmer <number>", "Set the reflections trimmer. reduces refractions. Accepted values: 0.0-1.0");
	au.addCommandLineOption("--thread-count <number>", "Set the number of working threads. Auto-detection according to CPU core count if not set. Accepted values: 0+");
	au.addCommandLineOption("--recursion-depth <number>", "Set the maximum recursion depth. Default: 3. Accepted values: 0+");
	au.addCommandLineOption("[--no-phong | --flat-shading]", "Disable collision normal phong (per-vertex) interpolation resulting in flat (per-face) shading.");
	au.addCommandLineOption("--no-textures", "Disable textures.");
	au.addCommandLineOption("--no-bilinear-filtering", "Disable bilinear texture filtering.");
	au.addCommandLineOption("--no-recursion", "Disable recursion. Same as setting recursion depth to 0.");
	au.addCommandLineOption("--no-reflections", "Disable reflections.");
	au.addCommandLineOption("--no-refractions", "Disable refractions.");
	au.addCommandLineOption("--no-self-shadows", "Disable self shadows.");
	au.addCommandLineOption("--export-scene <filename>", "Save the visualized scene to <filename>, for debugging purposes.");



	unsigned int helpType = 0;
	if ((helpType = argumentParser->readHelpType())) {
		argumentParser->getApplicationUsage()->write(std::cout, helpType);
		_error = 1;
		return;
	}

	// report any errors if they have occurred when parsing the program arguments.
	if (argumentParser->errors()) {
		argumentParser->writeErrorMessages(std::cout);
		_error = 10;
		return;
	}

	if (argumentParser->argc() <= 1) {
		argumentParser->getApplicationUsage()->write(std::cout, osg::ApplicationUsage::KEYBOARD_MOUSE_BINDING);
		_error = 10;
		return;
	}



	// read options
	std::string tmp, tmp2;

	while (argumentParser->read("--console")) {
		_state.interactive = false;
	}


	while (argumentParser->read("--no-phong")) {
		_state.phong = false;
	}
	while (argumentParser->read("--flat-shading")) {
		_state.phong = false;
	}
	while (argumentParser->read("--no-textures")) {
		_state.textures = false;
	}
	while (argumentParser->read("--no-bilinear-filtering")) {
		_state.bilinearFiltering = false;
	}
	while (argumentParser->read("--no-recursion")) {
		_state.recursion = false;
	}
	while (argumentParser->read("--no-reflections")) {
		_state.reflections = false;
	}
	while (argumentParser->read("--no-refractions")) {
		_state.refractions = false;
	}
	while (argumentParser->read("--no-self-shadows")) {
		_state.selfShadows = false;
	}


	while (argumentParser->read("--reflection-trimmer", tmp)) {
		_state.reflectionTrimmer = atof(tmp.c_str());
	}
	while (argumentParser->read("--refraction-trimmer", tmp)) {
		_state.refractionTrimmer = atof(tmp.c_str());
	}
	while (argumentParser->read("--recursion-depth", tmp)) {
		_state.recursionDepth = atoi(tmp.c_str());
	}
	while (argumentParser->read("--samples", tmp)) {
		_state.samplesPerPixel = atoi(tmp.c_str());
	}
	while (argumentParser->read("--thread-count", tmp)) {
		_state.threadCount = atoi(tmp.c_str());
	}

	while (argumentParser->read("-o", _state.outputPath)) {
	}
	while (argumentParser->read("--output-path", _state.outputPath)) {
	}


	//	{
	//		std::string resolution = "";
	//		if (!argumentParser.read("-r", resolution)) {
	//			argumentParser.read("--resolution", resolution);
	//		}
	//		uint pos = resolution.find('x', 0);
	//		if (pos != std::string::npos) {
	//			std::string w_s = resolution.substr(0, pos + 1);
	//			std::string h_s = resolution.substr(pos + 1);
	//			_state.screenWidth = atoi(w_s.c_str());
	//			_state.screenHeight = atoi(h_s.c_str());
	//			if (!_state.screenWidth || !_state.screenHeight) {
	//				_state.screenHeight = _state.screenWidth = 0;
	//			}
	//		}
	//	}

	while (argumentParser->read("-r", tmp, tmp2)) {
	}
	while (argumentParser->read("--resolution", tmp, tmp2)) {
	}
	_state.screenWidth = atoi(tmp.c_str());
	_state.screenHeight = atoi(tmp2.c_str());

	if (argumentParser->read("--camera", tmp)) {
		if (tmp != "") {
			_state.initialCamera = dynamic_cast<osg::Camera*> (osgDB::readNodeFile(tmp));

			if (_state.initialCamera.get() == NULL) {
				std::cout << "nulova kamera" << std::endl;
			}
		}
	}

	tmp = "";

	while (argumentParser->read("--export-model", tmp)) {
	};

	osgDB::Registry::instance()->setBuildKdTreesHint(osgDB::ReaderWriter::Options::BUILD_KDTREES);
	_state.scene = osgDB::readNodeFiles(*argumentParser);
	if (!_state.scene) {
		_error = 2;
	}

	if (tmp != "") {
		osgDB::writeNodeFile(*(_state.scene.get()), _state.outputPath + tmp);
		//		osgDB::writeNodeFile(*(_state.scene), tmp);
	}

	// correct possible bad values
	_state.reflectionTrimmer = NONNEGATIVE(CLAMP(_state.reflectionTrimmer));
	_state.refractionTrimmer = NONNEGATIVE(CLAMP(_state.refractionTrimmer));
	_state.secondaryRayAttenuation_constant = NONNEGATIVE(_state.secondaryRayAttenuation_constant);
	_state.secondaryRayAttenuation_linear = NONNEGATIVE(_state.secondaryRayAttenuation_linear);
	_state.secondaryRayAttenuation_quadratic = NONNEGATIVE(_state.secondaryRayAttenuation_quadratic);




	//	// any option left unread are converted into errors to write out later.
	//	argumentParser.reportRemainingOptionsAsUnrecognized();
	//	// report any errors if they have occurred when parsing the program arguments.
	//	if (argumentParser.errors()) {
	//		argumentParser.writeErrorMessages(std::cerr);
	//	}
}


void UILayer::setDefaults() {
	_state.interactive = true;

	_state.phong = true;
	_state.textures = true;
	_state.bilinearFiltering = true;
	_state.recursion = true;
	_state.reflections = true;
	_state.refractions = true;
	_state.selfShadows = true;
	_state.reflectionTrimmer = 1.;
	_state.refractionTrimmer = 1.;
	_state.secondaryRayAttenuation_constant = 1.0;
	_state.secondaryRayAttenuation_linear = 0.8;
	_state.secondaryRayAttenuation_quadratic = 0.0;

	_state.recursionDepth = 3;
	_state.samplesPerPixel = 1;
	_state.threadCount = 0;

	_state.screenWidth = 0;
	_state.screenHeight = 0;
	_state.initialCamera = NULL;
	_state.scene = NULL;
}


UILayer* UILayer::setRayManager(RayManager* rm) {
	_rm = rm;
	return this;
}


UILayer* UILayer::setWindowManager(WindowManager* wm) {
	_wm = wm;
	//	bindWM();
	return this;
}


WindowManager* UILayer::getWindowManager() const {
	return _wm;
}


RayManager* UILayer::getRayManager() const {
	return _rm;
}


osg::ArgumentParser& UILayer::getArguments() const {
	if (!_arguments) {
		std::cout << "Error: no arguments." << std::endl;
		return *(new osg::ArgumentParser(0, NULL));
	}
	return *_arguments;
}


bool UILayer::configureRM() {
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return 0;
	}

	_rm->setRecursionDepth(_state.recursionDepth);
	_rm->setThreadCount(_state.threadCount);
	_rm->enableTextures(_state.textures);
	_rm->enablePhongShading(_state.phong);
	_rm->enableBilinearFiltering(_state.bilinearFiltering);
	_rm->enableRecursion(_state.recursion);
	_rm->enableReflections(_state.reflections);
	_rm->enableRefractions(_state.refractions);
	_rm->enableSelfShadows(_state.selfShadows);
	_rm->setSamplesPerPixel(_state.samplesPerPixel);

	return true;
}


uchar UILayer::getProgress() const {
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return 0;
	}
	return _rm->getProgress();
}


bool UILayer::setRecursionDepth(uchar x) {
	_state.recursionDepth = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Recursion depth: " << (int) (_state.recursionDepth) << std::endl;
	_rm->setRecursionDepth(x);
	return true;
}


bool UILayer::setThreadCount(ushort x) {
	_state.threadCount = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Thread count: ";
	if (_state.threadCount > 0) {
		std::cout << (int) _state.threadCount;
	} else {
		std::cout << "AUTO";
	}
	std::cout << std::endl;
	_rm->setThreadCount(x);
	return true;
}


bool UILayer::enableTextures(bool x) {
	_state.textures = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Textures: " << (_state.textures ? "ON" : "OFF") << std::endl;
	_rm->enableTextures(x);
	return true;
}


bool UILayer::enablePhongShading(bool x) {
	_state.phong = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Phong shading: " << (_state.phong ? "ON" : "OFF") << std::endl;
	_rm->enablePhongShading(x);
	return true;
}


bool UILayer::enableBilinearFiltering(bool x) {
	_state.bilinearFiltering = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Bilinear texture filtering: " << (_state.bilinearFiltering ? "ON" : "OFF") << std::endl;
	_rm->enableBilinearFiltering(x);
	return true;
}


bool UILayer::enableRecursion(bool x) {
	_state.recursion = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Recursion: " << (_state.recursion ? "ON" : "OFF") << std::endl;
	_rm->enableRecursion(x);
	return true;
}


bool UILayer::enableReflections(bool x) {
	_state.reflections = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Reflections: " << (_state.reflections ? "ON" : "OFF") << std::endl;
	_rm->enableReflections(x);
	return true;
}


bool UILayer::enableRefractions(bool x) {
	_state.refractions = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Refractions: " << (_state.refractions ? "ON" : "OFF") << std::endl;
	_rm->enableRefractions(x);
	return true;
}


bool UILayer::enableSelfShadows(bool x) {
	_state.selfShadows = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	std::cout << "Self shadows: " << (_state.selfShadows ? "ON" : "OFF") << std::endl;
	_rm->enableSelfShadows(x);
	return true;
}


bool UILayer::setMultisampling(uchar samples) {
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	if (samples == 4 || samples == 8) {
		std::cout << "Multisampling: " << (int) samples << "x" << std::endl;
		_state.samplesPerPixel = samples;
	} else {
		std::cout << "Multisampling: OFF" << std::endl;
		_state.samplesPerPixel = 1;
	}
	_rm->setSamplesPerPixel(_state.samplesPerPixel);
	return true;
}


bool UILayer::setChunkSize(uint x) {
	_state.chunkSize = x;
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	_rm->setChunkSize(x);
	return true;
}


bool UILayer::setTrimmer(uchar trimmer, double value) {
	if (!_rm) {
		std::cout << "Error: no manager set." << std::endl;
		return false;
	}
	if (trimmer == 0 || trimmer == 1) {
		_state.reflectionTrimmer = value;
		std::cout << "Trimmer/reflections: " << _state.reflectionTrimmer << std::endl;
	}
	if (trimmer == 0 || trimmer == 2) {
		_state.refractionTrimmer = value;
		std::cout << "Trimmer/refractions: " << _state.reflectionTrimmer << std::endl;
	}
	_rm->setTrimmer(trimmer, value);
	return true;
}






///**
// * @note: not used.
// */
//void UILayer::bindWM() {
//	_wm->addEventHandler(this, WindowManager::VIEW_BOTH);
//}

//
//bool UILayer::handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa) {
//	if (!ea.getHandled() && ea.getEventType() == osgGA::GUIEventAdapter::KEYDOWN) {
//		int key = ea.getKey();
//
//		uint t = time(0);
//		switch (key) {
//			case 'r':
//				_wm->prepareBuffers();
//				// manazer pre okna: otvor okno pre raytracing
//				if (_rm->isActive()) {
//					_wm->closeRayWindow();
//				} else {
//					_wm->openRayWindow();
//				}
//
//				// manazer pre ray tracing: vytvor snimku
//				_rm->requestNewFrame();
//				return true;
//				break;
//			case 'p':
//				if (!_rm->isPointTracing()) {
//					// manazer pre okna: otvor okno pre raytracing
//					_wm->openRayWindow();
//
//					// manazer pre ray tracing: vytvor snimku
//					_rm->startPointTracing();
//				} else {
//					_rm->stopPointTracing();
//				}
//				return true;
//				break;
//			case 'o':
//			case 'c':
//				_rm->writeOutput(_wm->getOutputPath(), t);
//				_rm->writeSnapshot(_wm->getOutputPath(), t);
//				// no break
//			case 'x':
//				_rm->storeMasterCamera(_wm->getOutputPath(), t);
//				return true;
//				break;
//			case osgGA::GUIEventAdapter::KEY_Escape:
//				//				if (!_rm->isActive()) {
//				//					_wm->getViewer()->setDone(true);
//				//				} else {
//				//					_rm->writeOutput();
//				//				}
//				return true;
//			default:
//				break;
//		}
//	} else if (!ea.getHandled() && ea.getEventType() == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
//		if (_rm->isPointTracing()) {
//			_rm->tracePoint(ea.getX(), ea.getY());
//			return true;
//		}
//	}
//
//	if (!ea.getHandled() && ea.getEventType() == osgGA::GUIEventAdapter::CLOSE_WINDOW) {
//		osgViewer::View *view = dynamic_cast<osgViewer::View *> (&aa);
//		if (view == _wm->getRayView()) {
//			// ulozit stav okna
//			_wm->saveRayViewState();
//		} else if (view == _wm->getSceneView()) {
//			// ukoncit aplikaciu
//			_wm->getViewer()->setDone(true);
//		}
//		//		_wm->getViewer()->setDone(true);
//	}
//
//	if (!ea.getHandled() && ea.getEventType() == osgGA::GUIEventAdapter::RESIZE) {
//
//		int new_w, new_h;
//
//		new_w = ea.getWindowWidth();
//		new_h = ea.getWindowHeight();
//
//		if (new_w != _wm->getSceneWidth() || new_h != _wm->getSceneHeight()) {
//			_wm->resize(new_w, new_h);
//			//			std::cout << ea.getWindowWidth() << "x" << ea.getWindowHeight() << " ";
//		}
//
//
//
//	}
//
//
//	return false;
//}
