/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/



#include <osgViewer/Viewer>
#include <osgViewer/CompositeViewer>
#include <osgGA/TrackballManipulator>
//#include <osgGA/FlightManipulator>
//#include <osgGA/DriveManipulator>
#include <osgGA/KeySwitchMatrixManipulator>
#include <osgGA/StateSetManipulator>
#include <osgGA/AnimationPathManipulator>
#include <osgGA/TerrainManipulator>
#include <osg/Texture2D>
#include <osgViewer/ViewerEventHandlers>
#include <osgDB/WriteFile>
//#include <osgGA/SphericalManipulator>
#include <osg/Texture>
#include <osg/Texture2D>
#include <osg/TextureRectangle>
#include <osg/TextureCubeMap>

#include "WindowManager.h"
#include "manipulators.h"
#include "UILayer.h"

// field of view
#define FOV 30.0f



namespace Rayme {


	class mKeySwitchManipulator : public osgGA::KeySwitchMatrixManipulator {


		/** Get the keyboard and mouse usage of this manipulator.*/
		virtual void getUsage(osg::ApplicationUsage& usage) const {
			usage.addKeyboardMouseBinding("[space]", "reset camera to home position");
			usage.addKeyboardMouseBinding("u", "Trackball camera manipulator");
			usage.addKeyboardMouseBinding("i", "Terrain camera manipulator");
			usage.addKeyboardMouseBinding("o", "FirstPerson camera manipulator");
		}

	};


	class mWindowSizeHandler : public osgViewer::WindowSizeHandler {


		/** Get the keyboard and mouse usage of this manipulator.*/
		virtual void getUsage(osg::ApplicationUsage& usage) const {
			usage.addKeyboardMouseBinding("f", "toggle full-screen mode");
		}

	};


	class mStatsHandler : public osgViewer::StatsHandler {


		/** Get the keyboard and mouse usage of this manipulator.*/
		virtual void getUsage(osg::ApplicationUsage& usage) const {
			usage.addKeyboardMouseBinding("s", "toggle on-screen statistics");
		}

	};


	class mStateSetManipulator : public osgGA::StateSetManipulator {


		/** Get the keyboard and mouse usage of this manipulator.*/
		virtual void getUsage(osg::ApplicationUsage& usage) const {
			usage.addKeyboardMouseBinding(reinterpret_cast<const char*> (&_keyEventToggleTexturing), "Toggle texturing");
			usage.addKeyboardMouseBinding(reinterpret_cast<const char*> (&_keyEventCyclePolygonMode), "Toggle polygon fill mode between fill, line (wire frame) and points");
		}


		bool handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa) {
			if (!_stateset.valid()) return false;

			if (!_initialized) {
				_initialized = true;
				_backface = (_stateset->getMode(GL_CULL_FACE) & osg::StateAttribute::ON);
				_lighting = (_stateset->getMode(GL_LIGHTING) & osg::StateAttribute::ON);

				unsigned int mode = osg::StateAttribute::INHERIT | osg::StateAttribute::ON;

				_texture = (_stateset->getTextureMode(0, GL_TEXTURE_2D) & mode) != 0 ||
					  (_stateset->getTextureMode(0, GL_TEXTURE_3D) & mode) != 0 ||
					  (_stateset->getTextureMode(0, GL_TEXTURE_RECTANGLE) & mode) != 0 ||
					  (_stateset->getTextureMode(0, GL_TEXTURE_CUBE_MAP) & mode) != 0;

#if !defined(OSG_GLES1_AVAILABLE) && !defined(OSG_GLES2_AVAILABLE)
				_texture |= ((_stateset->getTextureMode(0, GL_TEXTURE_1D) & mode) != 0);
#endif
			}

			if (ea.getHandled()) return false;

			if (ea.getEventType() == osgGA::GUIEventAdapter::KEYDOWN) {
				if (ea.getKey() == _keyEventToggleTexturing) {
					setTextureEnabled(!getTextureEnabled());
					aa.requestRedraw();
					//return true;
				}
				if (ea.getKey() == _keyEventCyclePolygonMode) {
					cyclePolygonMode();
					aa.requestRedraw();
					return true;
				}
			}

			return false;
		}

	};


}


using namespace Rayme;


WindowManager::WindowManager(UILayer* ui) {
	_viewer = new osgViewer::CompositeViewer(ui->getArguments());
	_sceneView = new osgViewer::View;
	_rayView = new osgViewer::View;
	_sceneView->setName("real-time scene view");
	_rayView->setName("ray tracer view");
	_viewer->addView(_sceneView);
	_viewer->addView(_rayView);
	_raySavedTraits = NULL;
	_resizeRayWindow = false;

	// set up the camera manipulators and event handlers
	this->setManipulatorsAndEventHandlers(_viewer, ui->getArguments());
	//	this->addEventHandler(ui, WindowManager::VIEW_BOTH); //TODO preco toto duplikuje UILayer ?? (dvakrat sa vola destruktor)

	// set the threading model
	_viewer->setThreadingModel(osgViewer::ViewerBase::ThreadPerContext);

	// set the viewer to only refresh when needed (saves computation time for the working threads)
	_viewer->setRunFrameScheme(osgViewer::ViewerBase::ON_DEMAND);

	// disable the default close-on-esc-key behaviour
	_viewer->setKeyEventSetsDone(0);

	// get the output path
	_outputPath = ui->getUIState().outputPath;

	// get the resolution of the primary screen
	osg::GraphicsContext::WindowingSystemInterface* wsi = osg::GraphicsContext::getWindowingSystemInterface();
	unsigned screenWidth, screenHeight;
	wsi->getScreenResolution(osg::GraphicsContext::ScreenIdentifier(0), screenWidth, screenHeight);

	// get prefered initial window resolution
	int defaultWindowWidth = ui->getUIState().screenWidth, defaultWindowHeight = ui->getUIState().screenHeight;

	//	defaultWindowWidth = (defaultWindowWidth == 0 ? 800 : defaultWindowWidth);
	//	defaultWindowHeight = (defaultWindowHeight == 0 ? 600 : defaultWindowHeight);

	if (defaultWindowWidth > 0 && defaultWindowHeight > 0) {
		_sceneWidth = defaultWindowWidth;
		_sceneHeight = defaultWindowHeight;
	} else {
		_sceneWidth = screenWidth / 2 - 50;
		_sceneHeight = screenHeight / 2;
	}


	// sceneView
	{
		osg::ref_ptr<osg::GraphicsContext::Traits> traits = new osg::GraphicsContext::Traits;
		traits->x = (screenWidth - _sceneWidth <= 0) ? 40 : (screenWidth - _sceneWidth) / 2;
		traits->y = (screenHeight - _sceneHeight <= 0) ? 40 : (screenHeight - _sceneWidth) / 2;
		traits->width = _sceneWidth;
		traits->height = _sceneHeight;
		traits->windowDecoration = true;
		traits->doubleBuffer = true;
		traits->sharedContext = 0;
		traits->windowName = "Rayme - scene window";
		traits->samples = 4; // multisampling (antialias 4x)
		osg::ref_ptr<osg::GraphicsContext> gc = osg::GraphicsContext::createGraphicsContext(traits.get());
		osg::ref_ptr<osg::Camera> camera = _sceneView->getCamera();
		camera->setGraphicsContext(gc.get());
		camera->setViewport(new osg::Viewport(0, 0, _sceneWidth, _sceneHeight));

		camera->setProjectionMatrixAsPerspective(FOV, static_cast<double> (_sceneWidth) / static_cast<double> (_sceneHeight), 1.0f, 10000.0f);

		// ?
		GLenum buffer = traits->doubleBuffer ? GL_BACK : GL_FRONT;
		camera->setDrawBuffer(buffer);
		camera->setReadBuffer(buffer);

	}

	// rayView
	{
		osg::ref_ptr<osg::Camera> camera = _rayView->getCamera();
		//		osg::ref_ptr<osg::GraphicsContext::Traits> traits = new osg::GraphicsContext::Traits;
		//		traits->x = windowWidth + 40;
		//		traits->y = 40;
		//		traits->width = windowWidth;
		//		traits->height = windowHeight;
		//		traits->windowDecoration = true;
		//		traits->doubleBuffer = true;
		//		traits->sharedContext = 0;
		//		traits->supportsResize = false;
		//		traits->windowName = "Rayme - ray tracer window";
		//		osg::ref_ptr<osg::GraphicsContext> gc = osg::GraphicsContext::createGraphicsContext(traits.get());
		//
		//		camera->setGraphicsContext(gc.get());
		//		// ?
		//		GLenum buffer = traits->doubleBuffer ? GL_BACK : GL_FRONT;
		//		camera->setDrawBuffer(buffer);
		//		camera->setReadBuffer(buffer);

		camera->setComputeNearFarMode(osgUtil::CullVisitor::DO_NOT_COMPUTE_NEAR_FAR);
		//		camera->setViewport(new osg::Viewport(0, 0, traits->width, traits->height));
		//		camera->setAllowEventFocus(false);


		// turn lighting OFF
		// vypnut osvetlenie
		camera->getOrCreateStateSet()->setMode(GL_LIGHTING, osg::StateAttribute::OFF);


		// GC ?? TODO
		//void 	bindPBufferToTexture (GLenum buffer)



		_sceneView->setSceneData(ui->getUIState().scene.get());
		//		_sceneView->setSceneData(ui->getUIState().scene);

	}



	//	viewer->setCameraWithFocus(sceneView->getCamera());

	//	osgViewer::GraphicsWindow method: 
	//
	// /** Set the window's position and size.*/ 
	// void setWindowRectangle(int x, int y, int width, int height)


	/**
	 * framebuffer-bound texture
	 */
	_sceneImage = new osg::Image;
	_sceneImage->allocateImage(_sceneWidth, _sceneHeight, 1, IMAGE_FORMAT, IMAGE_TYPE);
	//	_sceneImage->setPixelBufferObject(new osg::PixelBufferObject(_sceneImage));
	_sceneView->getCamera()->attach(osg::Camera::COLOR_BUFFER, _sceneImage, 1, 1);

	_sceneTexture = new osg::Texture2D;
	_sceneTexture->setTextureSize(_sceneWidth, _sceneHeight);
	_sceneTexture->setInternalFormat(IMAGE_FORMAT);
	_sceneTexture->setImage(_sceneImage);
	_sceneTexture->setResizeNonPowerOfTwoHint(false); // proti scalovaniu textury na mocninu dvoch


	/**
	 * texture used for ray tracing output
	 */
	_rayTexture = new osg::Texture2D;
	_rayImage = new osg::Image;
	_rayImage->allocateImage(_sceneWidth, _sceneHeight, 1, IMAGE_FORMAT, IMAGE_TYPE);
	_rayImage->setPixelBufferObject(new osg::PixelBufferObject(_rayImage));
	_rayTexture->setImage(_rayImage);
	_rayTexture->setResizeNonPowerOfTwoHint(false);
	//	osg::Group *bufferNode = new osg::Group;
	//	bufferNode->addChild(createPolygonTextureScreen(_rayTexture));
	//	_rayView->setSceneData(bufferNode);

	_rayView->setSceneData(createPolygonTextureScreen(_rayTexture));


	/**
	 * cameras setup
	 */
	osg::Camera *cam = _rayView->getCamera();
	cam->setRenderOrder(osg::Camera::POST_RENDER, 2); // !
	cam->setProjectionMatrix(osg::Matrix::identity());
	cam->setViewMatrix(osg::Matrix::identity());
	cam->setViewport(0, 0, _sceneWidth, _sceneHeight);

	cam = _sceneView->getCamera();
	cam->setRenderOrder(osg::Camera::PRE_RENDER, 1); // !

}


WindowManager::~WindowManager() {
	if (_raySavedTraits != NULL) {
		delete(_raySavedTraits);
	}
}


void WindowManager::prepareBuffers() {

	resizeRayBuffers(_sceneWidth, _sceneHeight);
	_rayView->setSceneData(createPolygonTextureScreen(_rayTexture));
	_rayView->getCamera()->setViewport(0, 0, _sceneWidth, _sceneHeight);

	//	osg::Geode *geode = dynamic_cast<osg::Geode *> (_rayView->getScene());
	//	osg::Drawable *dr = geode ? geode->getDrawable(1) : NULL;
	//	osg::Geometry *geometry = dr ? dr->asGeometry() : NULL;
	//	osg::StateSet* stateset = geometry ? geometry->getStateSet() : NULL;
	//	if (stateset) {
	//		stateset->setTextureAttributeAndModes(0, _rayTexture, osg::StateAttribute::ON);
	//	} else {
	//		std::cout << "puca drevena ";
	//	}
}


void WindowManager::resize(unsigned int width, unsigned int height) {
	_sceneWidth = width;
	_sceneHeight = height;

	this->resizeSceneBuffers(width, height);
	_resizeRayWindow = true;

	//	osg::Camera *cam = _rayView->getCamera();
	//	cam->setViewport(0, 0, _sceneWidth, _sceneHeight);

}


void WindowManager::resizeSceneBuffers(unsigned int width, unsigned int height) {
	// framebuffer-bound texture
	_sceneTexture->setTextureSize(width, height);
	_sceneView->getCamera()->detach(osg::Camera::COLOR_BUFFER);
	//	_sceneImage = new osg::Image;
	_sceneImage->allocateImage(width, height, 1, IMAGE_FORMAT, IMAGE_TYPE);
	_sceneView->getCamera()->attach(osg::Camera::COLOR_BUFFER, _sceneImage, 1, 1);
	_sceneTexture->setImage(0, _sceneImage);

	//	std::cout << width << "x" << height << " ";
}


void WindowManager::resizeRayBuffers(unsigned int width, unsigned int height) {
	//	_rayImage = new osg::Image;
	_rayImage->allocateImage(width, height, 1, IMAGE_FORMAT, IMAGE_TYPE);
	//	_rayTexture->setImage(0, _rayImage);
	// texture used for ray tracing output
	_rayTexture->setTextureSize(width, height);

	//	_rayImage->setPixelBufferObject(new osg::PixelBufferObject(_rayImage));

	//	std::cout << width << "x" << height << " ";
}


//void WindowManager::adjustViewports(unsigned int width, unsigned int height) {
//}


/* 
 * TODO clean up, omg !!!
 */
void WindowManager::setManipulatorsAndEventHandlers(osgViewer::CompositeViewer *cv, osg::ArgumentParser &arguments) {
	osg::ref_ptr<mKeySwitchManipulator> keyswitchManipulator = new mKeySwitchManipulator; // keyswitch manipulator
	osg::ref_ptr<mStateSetManipulator> statesetManipulator = new mStateSetManipulator; // state manipulator
	//	osg::ref_ptr<osgViewer::ThreadingHandler> th = new osgViewer::ThreadingHandler; // thread model handler
	osg::ref_ptr<mWindowSizeHandler> wsh = new mWindowSizeHandler; // window size toggle handler
	osg::ref_ptr<mStatsHandler> sh = new mStatsHandler; // stats handler
	osg::ref_ptr<osgViewer::HelpHandler> hh = new osgViewer::HelpHandler(arguments.getApplicationUsage()); // help handler
	//	osg::ref_ptr<osgViewer::RecordCameraPathHandler> rcph = new osgViewer::RecordCameraPathHandler; // record camera path handler
	//	osg::ref_ptr<osgViewer::LODScaleHandler> lodh = new osgViewer::LODScaleHandler; // LOD Scale handler
	//	osg::ref_ptr<osgViewer::ScreenCaptureHandler> sch = new osgViewer::ScreenCaptureHandler; // screen capture handler

	keyswitchManipulator->addMatrixManipulator('u', "Trackball", new osgGA::TrackballManipulator());
	keyswitchManipulator->addMatrixManipulator('i', "Terrain", new osgGA::TerrainManipulator());
	keyswitchManipulator->addMatrixManipulator('o', "FirstPerson", new FPSVerticalManipulator());

	//        keyswitchManipulator->addMatrixManipulator( '7', "Spherical", new osgGA::SphericalManipulator() );
	//        keyswitchManipulator->addMatrixManipulator( '2', "Flight", new osgGA::FlightManipulator() );
	//		keyswitchManipulator->addMatrixManipulator('3', "Drive", new osgGA::DriveManipulator());
	//        keyswitchManipulator->addMatrixManipulator( '5', "Orbit", new osgGA::OrbitManipulator() );

	//	std::string pathfile;
	//	double animationSpeed = 1.0;
	//	while (arguments.read("--speed", animationSpeed)) {
	//	}
	//	char keyForAnimationPath = '8';
	//	while (arguments.read("-p", pathfile)) {
	//		osgGA::AnimationPathManipulator* apm = new osgGA::AnimationPathManipulator(pathfile);
	//		if (apm || !apm->valid()) {
	//			apm->setTimeScale(animationSpeed);
	//
	//			unsigned int num = keyswitchManipulator->getNumMatrixManipulators();
	//			keyswitchManipulator->addMatrixManipulator(keyForAnimationPath, "Path", apm);
	//			keyswitchManipulator->selectMatrixManipulator(num);
	//			++keyForAnimationPath;
	//		}
	//	}

	std::vector<osgViewer::View*> views;
	cv->getViews(views, true);
	if (views.size() < 1) {
		return;
	}

	statesetManipulator->setStateSet(cv->getView(0)->getCamera()->getOrCreateStateSet());
	hh->setApplicationUsage(arguments.getApplicationUsage());

	int i = 0;
	views[i]->setCameraManipulator(keyswitchManipulator.get());
	views[i]->addEventHandler(statesetManipulator.get());
	//	views[i]->addEventHandler(th.get());
	views[i]->addEventHandler(sh.get());
	views[i]->addEventHandler(hh.get());
	views[i]->addEventHandler(wsh.get());
	//	views[i]->addEventHandler(rcph.get());
	//	views[i]->addEventHandler(lodh.get());
	//	views[i]->addEventHandler(sch.get());


	views[1]->setCameraManipulator(new DiscardManipulator);

	//	for (std::vector<osgViewer::View*>::iterator it = views.begin(); it != views.end(); ++it) {
	//		(*it)->addEventHandler(wsh.get());
	//	}

	//		for (std::vector<osgViewer::View*>::iterator it = views.begin(); it != views.end(); ++it) {
	//			(*it)->setCameraManipulator(keyswitchManipulator.get());
	//			(*it)->addEventHandler(statesetManipulator.get());
	//			(*it)->addEventHandler(th.get());
	//			(*it)->addEventHandler(wsh.get());
	//			(*it)->addEventHandler(sh.get());
	//			(*it)->addEventHandler(hh.get());
	//			(*it)->addEventHandler(rcph.get());
	//			(*it)->addEventHandler(lodh.get());
	//			(*it)->addEventHandler(sch.get());
	//		}


	//	statesetManipulator->setStateSet(cv->getView(0)->getCamera()->getOrCreateStateSet());
	//	for (unsigned i = 0; i < views.size(); i++) {
	//		views[i]->setCameraManipulator(keyswitchManipulator.get());
	//		views[i]->addEventHandler(statesetManipulator.get());
	//		views[i]->addEventHandler(th.get());
	//			views[i]->addEventHandler(wsh.get());
	//		views[i]->addEventHandler(sh.get());
	//		views[i]->addEventHandler(hh.get());
	//		views[i]->addEventHandler(rcph.get());
	//		views[i]->addEventHandler(lodh.get());
	//		views[i]->addEventHandler(sch.get());
	//	}
}


void WindowManager::addEventHandler(osgGA::GUIEventHandler* eventHandler, WindowManager::SelectView select) {
	if (select & WindowManager::VIEW_SCENE) {
		_sceneView->addEventHandler(eventHandler);
	}
	if (select & WindowManager::VIEW_RAY) {
		_rayView->addEventHandler(eventHandler);
	}
}


osg::Node* WindowManager::createPolygonTextureScreen(osg::Texture2D *outputTexture) {
	osg::Geode *geode = new osg::Geode;
	//		bufferNode->addChild(geode);


	osg::Geometry* polyGeom = new osg::Geometry();

	polyGeom->setName("PolyGeom");

	polyGeom->setDataVariance(osg::Object::DYNAMIC);
	polyGeom->setSupportsDisplayList(false);

	osg::Vec3Array* vertices = new osg::Vec3Array;
	vertices->push_back(osg::Vec3(-1.0f, -1.0f, 0.0f));
	vertices->push_back(osg::Vec3(1.0f, -1.0f, 0.0f));
	vertices->push_back(osg::Vec3(1.0f, 1.0f, 0.0f));
	vertices->push_back(osg::Vec3(-1.0f, 1.0f, 0.0f));

	osg::Vec2Array* texcoords = new osg::Vec2Array;
	texcoords->push_back(osg::Vec2(0.0f, 0.0f));
	texcoords->push_back(osg::Vec2(1.0f, 0.0f));
	texcoords->push_back(osg::Vec2(1.0f, 1.0f));
	texcoords->push_back(osg::Vec2(0.0f, 1.0f));
	polyGeom->setVertexArray(vertices);
	polyGeom->setTexCoordArray(0, texcoords);

	osg::Vec4Array* colors = new osg::Vec4Array;
	colors->push_back(osg::Vec4(1.0f, 1.0f, 1.0f, 1.0f));
	polyGeom->setColorArray(colors);
	polyGeom->setColorBinding(osg::Geometry::BIND_OVERALL);

	polyGeom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::QUADS, 0, vertices->size()));

	// new we need to add the texture to the Drawable, we do so by creating a 
	// StateSet to contain the Texture StateAttribute.
	osg::StateSet* stateset = new osg::StateSet;

	stateset->setTextureAttributeAndModes(0, outputTexture, osg::StateAttribute::ON);

	polyGeom->setStateSet(stateset);

	geode->addDrawable(polyGeom);


	return geode;
}


void WindowManager::openRayWindow() {

	if (_resizeRayWindow) {
		resizeRayBuffers(_sceneWidth, _sceneHeight);
	}

	osg::GraphicsContext *gc = getRayView()->getCamera()->getGraphicsContext();
	osg::ref_ptr<osg::GraphicsContext::Traits> traits;

	osg::GraphicsContext::WindowingSystemInterface* wsi = osg::GraphicsContext::getWindowingSystemInterface();
	unsigned screenWidth, screenHeight;


	if (gc && gc->valid()) {
		closeRayWindow(); // this also saves tha current context state
	}

	if (!gc || !gc->valid()) {
		if (!_raySavedTraits) {
			wsi->getScreenResolution(osg::GraphicsContext::ScreenIdentifier(0), screenWidth, screenHeight);
			traits = new osg::GraphicsContext::Traits;
			//		traits->x = screenWidth - _sceneWidth - 40;
			//		traits->y = 40;
			traits->x = _sceneView->getCamera()->getGraphicsContext()->getTraits()->x + 100;
			traits->y = _sceneView->getCamera()->getGraphicsContext()->getTraits()->y + 100;
			traits->width = _sceneWidth;
			traits->height = _sceneHeight;
			traits->windowDecoration = true;
			traits->doubleBuffer = true;
			traits->sharedContext = 0;
			traits->supportsResize = false;
			traits->windowName = "Rayme - ray tracer window";
		} else {
			traits = _raySavedTraits;
			traits->width = _sceneWidth;
			traits->height = _sceneHeight;
		}


		//	osgViewer::ViewerBase::Windows ws;
		//	_rayView->getViewerBase()->getWindows(ws, true);
		//
		//	if (ws.size() > 1) {
		//		ws[1]->setWindowRectangle(10, 10, 100, 100);
		//	}

		osg::ref_ptr<osg::GraphicsContext> gc = osg::GraphicsContext::createGraphicsContext(traits.get());
		osg::ref_ptr<osg::Camera> camera = _rayView->getCamera();

		camera->setGraphicsContext(gc.get());
		// ?
		GLenum buffer = traits->doubleBuffer ? GL_BACK : GL_FRONT;
		camera->setDrawBuffer(buffer);
		camera->setReadBuffer(buffer);

		camera->setViewport(new osg::Viewport(0, 0, traits->width, traits->height));


		osg::Timer_t timer = _sceneView->getStartTick();
		_viewer->realize();
		_viewer->setStartTick(timer);
	} else {

		osgViewer::ViewerBase::Windows w;
		_rayView->getViewerBase()->getWindows(w, true);
		//		std::cout << "windows vector size: " << w.size() << std::endl;

		if (w.size() > 1) {
			const osg::GraphicsContext::Traits * traits = _rayView->getCamera()->getGraphicsContext()->getTraits();
			w[1]->setWindowRectangle(traits->x, traits->y, _sceneWidth, _sceneHeight);
		}

	}



}


void WindowManager::closeRayWindow() {
	osg::GraphicsContext *gc = getRayView()->getCamera()->getGraphicsContext();
	if (gc) {
		saveRayViewState();
		//		gc->close();//TODO toto serie aplikaciu!
	}
}


void WindowManager::saveRayViewState() {
	if (_raySavedTraits == NULL) {
		_raySavedTraits = new osg::GraphicsContext::GraphicsContext::Traits();
	}
	*(_raySavedTraits) = *(_rayView->getCamera()->getGraphicsContext()->getTraits());
}


int WindowManager::run() {
	_viewer->realize();
	return _viewer->run();
}


