/* ****************************************************************************
 *
 * rayme - ray tracing application for virtual 3D scenes 
 *
 * ----------------------------------------------------------------------------
 *
 * Andrej Rypak, xrypak@gmail.com, 2011 - 2012
 *
 * Thsesis project - Raytracing of Virtual Graphics Scenes
 * Brno University of Technology, Faculty of Information Technologies
 * Andrej Rypak, xrypak00@stud.fit.vutbr.cz, 2011/2012,
 * under supervision of Jan Peciva, PhD.
 *
 ******************************************************************************/


/**
 * main.cpp
 * 
 * The main application file containing the main() function.
 */


#include "common.h"


#include <osg/ArgumentParser>
#include <osg/Node>
#include <osg/Camera>
#include <osg/Image>
#include <osgUtil/Optimizer>
#include <osgDB/ReadFile>
#include <osgDB/WriteFile>
#include <osgViewer/CompositeViewer>

#include "WindowManager.h"
#include "RayManager.h"
#include "EventHandler.h"
#include "UILayer.h"

//#include <QtGui/QApplication>
//#include "widget.h"


using namespace Rayme;


int main(int argc, char** argv) {

	//    QApplication a(argc, argv);
	//	Widget w;
	//    w.show();
	//	a.exec();
	//	return 0;

	UILayer* ui = new UILayer(argc, argv);

	switch (ui->getError()) {
		case 0:
			// no error
			break;
		case 1:
			// help
			return 0;
		default:
			return 1;
	}


	//		osgUtil::Optimizer optimizer;
	//		optimizer.optimize(ui->getUIState().scene.get());
			osgUtil::Optimizer::TextureVisitor tv(true, false, false, false, false, false);
			ui->getUIState().scene->accept(tv);

	if (ui->getUIState().interactive) {
		// construct the ray tracing manager and bind the appropriate handlers to both managers
		WindowManager *wm = new WindowManager(ui);
		RayManager *rm = new RayManager(wm->getSceneView()->getCamera(), ui->getUIState().scene.get(), wm->getRayImage(), wm->getSceneImage(), wm->getRayView());
		//		RayManager *rm = new RayManager(wm->getSceneView()->getCamera(), ui->getUIState().scene, wm->getRayImage(), wm->getSceneImage(), wm->getRayView());
		wm->addEventHandler(new EventHandler(*ui, rm, wm), WindowManager::VIEW_BOTH);


		// optimize the scene graph, remove redundant nodes and state etc.
		// this is commented out for a reason - texture images get unrefed after applying to openGL
		//				osgUtil::Optimizer optimizer;
		//				optimizer.optimize(ui->getUIState().scene.get());
		//		osgUtil::Optimizer::TextureVisitor tv(true, false, false, false, false, false);
		//		ui->getUIState().scene->accept(tv);

		ui->setWindowManager(wm);
		ui->setRayManager(rm);
		ui->configureRM();

	
		//TODO this is not working ...
		//		if (camera.get()) {
		//			wm->getSceneView()->getCamera()->setProjectionMatrix(camera->getProjectionMatrix());
		//			wm->getSceneView()->getCamera()->setViewMatrix(camera->getViewMatrix());
		//			//			std::cout << "setting camera" << std::endl;
		//		}




		// any option left unread are converted into errors to write out later.
		ui->getArguments().reportRemainingOptionsAsUnrecognized();
		// report any errors if they have occurred when parsing the program arguments.
		if (ui->getArguments().errors()) {
			ui->getArguments().writeErrorMessages(std::cout);

			delete(rm);
			delete(wm);
			delete(ui);
			return 1;
		}

		int check = wm->run();

		delete(rm);
		delete(wm);
		delete(ui);

		return check;

	} else {
		// the program is running in console mode - no windows, just rendering
		osg::notify(osg::INFO) << ui->getArguments().getApplicationName() << ": console mode." << std::endl;

		// any option left unread are converted into errors to write out later.
		ui->getArguments().reportRemainingOptionsAsUnrecognized();
		// report any errors if they have occurred when parsing the program arguments.
		if (ui->getArguments().errors()) {
			ui->getArguments().writeErrorMessages(std::cout);
			return 1;
		}

		osg::ref_ptr<osg::Camera> cam;
		if (ui->getUIState().initialCamera.get()) {
			cam = ui->getUIState().initialCamera.get();
		} else {
			cam = new osg::Camera;
		}
		int defaultWindowWidth = ui->getUIState().screenWidth, defaultWindowHeight = ui->getUIState().screenHeight;

		if (defaultWindowWidth == 0 || defaultWindowHeight == 0) {
			// get resolution from viewport
			//			if (!ui->getUIState().initialCamera.get()) {
			if (!ui->getUIState().initialCamera) {
				std::cout << "Error: no resolution set, no camera defined." << std::endl;
				return 1;
			}
			defaultWindowWidth = ui->getUIState().initialCamera->getViewport()->width();
			defaultWindowHeight = ui->getUIState().initialCamera->getViewport()->height();
		} else {
			// set the viewport resolution
			cam->setViewport(0, 0, defaultWindowWidth, defaultWindowHeight);
		}

		osg::ref_ptr<osg::Image> image = new osg::Image;
		image->allocateImage(defaultWindowWidth, defaultWindowHeight, 1, IMAGE_FORMAT, IMAGE_TYPE);

		RayManager *rm = new RayManager(cam.get(), ui->getUIState().scene.get(), image, NULL, NULL);

		ui->setRayManager(rm);
		ui->configureRM();
		rm->setConsoleProgressPrintoutThreshold(5);

		rm->run(); // note: do not start a new thread as the app would end prematurely

		rm->writeOutput(ui->getUIState().outputPath);

		delete(rm);
		delete(ui);
	}



	return 0;
}


